use crate::{
    bootinfo::BOOTINFO,
    pit::NUM_INTERRUPTS,
    print, println,
    tasking::{InvalidPid, IpcMessage, SleepReason},
    virtual_memory::{ASpaceMutex, AddressSpace, PagingError, ACTIVE_SPACE, KERNEL_SPACE},
    TASKING,
};
use alloc::{boxed::Box, vec::Vec};
use az::WrappingCast;
use cast::{u64, usize};
use core::{
    arch::{asm, naked_asm},
    ptr, slice,
    sync::atomic::Ordering,
};
use hashbrown::HashMap;
use pic8259::ChainedPics;
use spin::{Lazy, Mutex, RwLock};
use x86_64::{
    registers::control::Cr2,
    set_general_handler,
    structures::{
        idt::{InterruptDescriptorTable, InterruptStackFrame, PageFaultErrorCode},
        paging::{
            mapper::TranslateResult, Page, PageTableFlags, PageTableIndex, PhysFrame, Translate,
        },
    },
    PhysAddr, PrivilegeLevel, VirtAddr,
};

#[cfg(feature = "log-rpc-pcapng")]
use crate::serial::SECOND_PORT;
#[cfg(feature = "log-rpc-pcapng")]
use saturating_cast::SaturatingCast;

#[cfg(feature = "log-rpc-text")]
use crate::ipc_dump::dump_ipc_message;

const IRQ_BASE: u8 = 32;

static IDT: Lazy<InterruptDescriptorTable> = Lazy::new(|| {
    let mut idt = InterruptDescriptorTable::new();
    set_general_handler!(&mut idt, general_handler);
    set_general_handler!(&mut idt, exc_handler, 0..32);
    set_general_handler!(&mut idt, irq_handler, 32..48);
    idt[0x80].set_handler_fn(syscall_handler_header).set_privilege_level(PrivilegeLevel::Ring3);
    idt.page_fault.set_handler_fn(page_fault_handler);
    idt.device_not_available.set_handler_fn(device_not_available_handler);
    idt
});
static PICS: Mutex<ChainedPics> = Mutex::new(unsafe { ChainedPics::new(IRQ_BASE, IRQ_BASE + 8) });
static IRQ_HANDLERS: RwLock<[Option<IrqHandler>; 16]> = RwLock::new([None; 16]);
static IRQ_TASKS: RwLock<[Option<usize>; 16]> = RwLock::new([None; 16]);

pub type IrqHandler = fn(irq_num: u8, eoi_guard: EoiGuard);

#[derive(Debug)]
pub struct InvalidIrq;

pub fn init() {
    IDT.load();
    unsafe { PICS.lock().initialize() };
    unsafe { PICS.lock().write_masks(0, 0) };
    x86_64::instructions::interrupts::enable();
}

extern "x86-interrupt" fn page_fault_handler(
    stack_frame: InterruptStackFrame,
    error_code: PageFaultErrorCode,
) {
    #[warn(clippy::expect_used, reason = "FIXME")]
    if let Ok(faulting_addr) = Cr2::read() {
        if faulting_addr.p4_index() >= PageTableIndex::new(256)
            && !ACTIVE_SPACE.lock().level_4_table()[faulting_addr.p4_index()]
                .flags()
                .contains(PageTableFlags::PRESENT)
            && KERNEL_SPACE.lock().level_4_table()[faulting_addr.p4_index()]
                .flags()
                .contains(PageTableFlags::PRESENT)
        {
            ACTIVE_SPACE.lock().level_4_table_mut()[faulting_addr.p4_index()] =
                KERNEL_SPACE.lock().level_4_table()[faulting_addr.p4_index()].clone();
            return;
        }
        if let Some(current_pid) = TASKING.current_pid() {
            print!("PID {current_pid} ");
        } else {
            print!("Kernel init ");
        }
        if error_code.contains(PageFaultErrorCode::PROTECTION_VIOLATION) {
            println!(
                "page faulted {error_code:#?} at {:#x}\nEntry flags: {:#?}\n{stack_frame:#?}",
                faulting_addr,
                match ACTIVE_SPACE.lock().translate(faulting_addr) {
                    TranslateResult::Mapped { flags, .. } => flags,
                    _ =>
                    {
                        #![allow(
                            clippy::panic,
                            reason = "A protection violation only happends on mapped addresses. If we get here, something has gone VERY wrong."
                        )]
                        panic!();
                    }
                },
            );
        } else {
            println!("page faulted {error_code:#?} at {:#x}\n{stack_frame:#?}", faulting_addr);
        }
    } else {
        if let Some(current_pid) = TASKING.current_pid() {
            print!("PID {current_pid} ");
        } else {
            print!("Kernel init ");
        }
        println!(
            "page faulted {error_code:#?} at non-caonical address {:#x}\n{stack_frame:#?}",
            Cr2::read_raw()
        );
    }
    TASKING.exit(254);
}

extern "x86-interrupt" fn device_not_available_handler(_stack_frame: InterruptStackFrame) {
    TASKING.switch_fpu_context();
}

#[expect(clippy::needless_pass_by_value, reason = "Signature dictated by external crate")]
fn general_handler(stack_frame: InterruptStackFrame, index: u8, _error_code: Option<u64>) {
    println!("Other interrupt {index}\n{stack_frame:#?}");
}

fn exc_handler(_stack_frame: InterruptStackFrame, _index: u8, _error_code: Option<u64>) {
    if let Some(current_pid) = TASKING.current_pid() {
        print!("PID {current_pid} ");
    } else {
        print!("Kernel init ");
    }
    println!("had exception, exiting");
    TASKING.exit(254);
}

pub struct EoiGuard(u8);

impl Drop for EoiGuard {
    fn drop(&mut self) {
        unsafe {
            PICS.lock().notify_end_of_interrupt(self.0);
        }
    }
}

#[derive(Copy, Clone, Debug)]
pub enum SendIpcError {
    InvalidPid,
    #[expect(unused, reason = "Useful in debug logs")]
    PagingError(PagingError),
}

impl From<InvalidPid> for SendIpcError {
    fn from(_v: InvalidPid) -> Self {
        Self::InvalidPid
    }
}

impl From<PagingError> for SendIpcError {
    fn from(v: PagingError) -> Self {
        Self::PagingError(v)
    }
}

pub fn send_ipc_to(
    pid: usize,
    buffer: Box<[u8], &'static ASpaceMutex>,
    len: usize,
    from_kernel: bool,
) -> Result<(), SendIpcError> {
    let from = if from_kernel {
        usize::MAX
    } else {
        #[expect(
            clippy::unwrap_used,
            reason = "from_kernel=false is used from the send_ipc syscall, which can only be called in usermode where PIDs are always present"
        )]
        TASKING.current_pid().unwrap()
    };
    #[cfg(feature = "log-rpc-text")]
    dump_ipc_message(from, pid, &buffer[0..len]);
    #[cfg(feature = "log-rpc-pcapng")]
    {
        #[expect(
            clippy::unwrap_used,
            reason = "The min call guarantees that the value is in the range of a u32 before the cast"
        )]
        let trunc_len: u32 = usize::min(len, 4096).try_into().unwrap();
        #[expect(
            clippy::arithmetic_side_effects,
            reason = "Can't underflow, as x % 4 < 4 no matter the x"
        )]
        let padding = if (trunc_len % 4) != 0 { 4 - (trunc_len % 4) } else { 0 };
        #[expect(
            clippy::arithmetic_side_effects,
            reason = "Can't overflow, as padding is no more than 4 and trunc_len is no more than 4096."
        )]
        let padded_len = trunc_len + padding;
        #[expect(
            clippy::arithmetic_side_effects,
            reason = "Can't overflow, as padded_len is no more than 4096 and 4096+24 < u32::MAX"
        )]
        let total_len = padded_len + 16 + (4 * 4);
        SECOND_PORT.write_u32s(&[
            0x3,                                                   // SPB type
            total_len,                                             // Total block length
            trunc_len.saturating_cast::<u32>().saturating_add(16), // Packet length
        ]);
        SECOND_PORT.write_bytes(&pid.to_ne_bytes());
        SECOND_PORT.write_bytes(&from.to_ne_bytes());
        #[expect(
            clippy::indexing_slicing,
            reason = "The truncated length is always <= the buffer's length"
        )]
        SECOND_PORT.write_bytes(&buffer[0..usize(trunc_len)]);
        for _ in 0..padding {
            SECOND_PORT.write_bytes(&[0]);
        }
        SECOND_PORT.write_u32s(&[
            total_len, // Total block length
        ]);
    }
    assert!(len <= buffer.len());
    TASKING.process_mut(pid, |process| {
        let buf_num_pages = buffer.len() / 4096;
        let buffer = Box::into_raw(buffer);
        #[expect(
            clippy::unwrap_used,
            reason = "ASpaceMutex is guaranteed to only return page-aligned allocations, so the address passed to Page::from_start_address is always valid"
        )]
        let buf_start_page = Page::from_start_address(VirtAddr::new(u64(buffer.expose_provenance()))).unwrap();
        // This is None only if the destiniation is the current process. If so,
        // no remapping is necessary so just return the old buffer.
        let dest_buffer = if let Some(aspace) = process.address_space_mut() {
            let page = ACTIVE_SPACE.lock().move_mappings_free(buf_start_page, buf_num_pages, aspace)?;
            ptr::slice_from_raw_parts_mut::<u8>(
                page.start_address().as_mut_ptr(),
                buffer.len(),
            )
        } else {
            buffer
        };
        let new_buffer_key = process.data_buffers().lock().insert(dest_buffer);
        let message = IpcMessage {
            buffer_idx: new_buffer_key,
            len,
            from,
        };
        process.message_queue().lock().push(message);
        Ok::<_, PagingError>(())
    })??;
    #[expect(
        clippy::unwrap_used,
        reason = "The PID is known valid due to using it in process_mut above"
    )]
    TASKING
        .process(pid, |process| {
            for (tid, _) in &*process.threads().read() {
                TASKING.wake(pid, tid, SleepReason::WaitingForIPC).unwrap();
            }
        })
        .unwrap();
    Ok(())
}

fn irq_handler(_stack_frame: InterruptStackFrame, index: u8, _error_code: Option<u64>) {
    #[expect(
        clippy::arithmetic_side_effects,
        reason = "This function is only called for irqs, which are always above the base index."
    )]
    let irq_num = index - IRQ_BASE;
    let eoi_guard = EoiGuard(index);
    #[expect(
        clippy::indexing_slicing,
        reason = "This function is only called for 16 irqs, which are all valid indexes"
    )]
    if let Some(handler) = IRQ_HANDLERS.read()[usize(irq_num)] {
        handler(irq_num, eoi_guard);
    } else if let Some(handler_task) = IRQ_TASKS.read()[usize(irq_num)] {
        let pid = handler_task;
        let mut buffer = Vec::new_in(&*ACTIVE_SPACE);
        buffer.extend_from_slice(&2u16.to_le_bytes());
        buffer.push(irq_num);
        let len = buffer.len();
        buffer.resize(len.next_multiple_of(4096), 0);
        let buffer = buffer.into_boxed_slice();
        match send_ipc_to(pid, buffer, len, true) {
            Ok(()) => (),
            Err(SendIpcError::InvalidPid) => {
                IRQ_TASKS.write()[usize(irq_num)] = None;
            }
            #[expect(
                clippy::panic,
                reason = "If an IRQ message fails to deliver, driver code can easify malfunction, so treat this as fatal."
            )]
            Err(e) => {
                panic!("Failed to send IRQ message: {e:?}");
            }
        }
    }
}

#[repr(C)]
#[derive(Debug, Clone, Copy)]
struct SyscallRegs {
    rax: u64,
    rbx: u64,
    rcx: u64,
    rdx: u64,
    rsi: u64,
    rdi: u64,
    rsp: u64,
    rbp: u64,
    r8: u64,
    r9: u64,
    r10: u64,
    r11: u64,
    r12: u64,
    r13: u64,
    r14: u64,
    r15: u64,
}

#[no_mangle]
static mut SYSCALL_REGS: SyscallRegs = SyscallRegs {
    rax: 0,
    rbx: 0,
    rcx: 0,
    rdx: 0,
    rsi: 0,
    rdi: 0,
    rsp: 0,
    rbp: 0,
    r8: 0,
    r9: 0,
    r10: 0,
    r11: 0,
    r12: 0,
    r13: 0,
    r14: 0,
    r15: 0,
};

#[naked]
extern "x86-interrupt" fn syscall_handler_header(stack_frame: InterruptStackFrame) {
    unsafe {
        naked_asm!(
            "mov [rip+SYSCALL_REGS], rax",
            "mov [rip+SYSCALL_REGS+8], rbx",
            "mov [rip+SYSCALL_REGS+16], rcx",
            "mov [rip+SYSCALL_REGS+24], rdx",
            "mov [rip+SYSCALL_REGS+32], rsi",
            "mov [rip+SYSCALL_REGS+40], rdi",
            "mov [rip+SYSCALL_REGS+48], rsp",
            "mov [rip+SYSCALL_REGS+56], rbp",
            "mov [rip+SYSCALL_REGS+64], r8",
            "mov [rip+SYSCALL_REGS+72], r9",
            "mov [rip+SYSCALL_REGS+80], r10",
            "mov [rip+SYSCALL_REGS+88], r11",
            "mov [rip+SYSCALL_REGS+96], r12",
            "mov [rip+SYSCALL_REGS+104], r13",
            "mov [rip+SYSCALL_REGS+112], r14",
            "mov [rip+SYSCALL_REGS+120], r15",
            "jmp syscall_handler",
        )
    }
}

fn get_buffer(id: u64) -> Option<Box<[u8], &'static ASpaceMutex>> {
    TASKING.current_process(|process| {
        process
            .data_buffers()
            .lock()
            .try_remove(usize(id))
            .map(|buf| unsafe { Box::from_raw_in(buf, &*ACTIVE_SPACE) })
    })
}

pub static REGISTERD_PIDS: Lazy<RwLock<HashMap<u64, u64>>> =
    Lazy::new(|| RwLock::new(HashMap::new()));

static INITRD_BUF: Lazy<&'static [u8]> = Lazy::new(|| {
    #[expect(clippy::expect_used, reason = "Boot cannot happen without an initrd")]
    let ramdisk_start = BOOTINFO.ramdisk_addr.into_option().expect("initrd not present");
    let ramdisk_len = BOOTINFO.ramdisk_len;
    let initrd = unsafe {
        slice::from_raw_parts(
            ptr::with_exposed_provenance::<u8>(usize(ramdisk_start)),
            usize(ramdisk_len),
        )
    };
    let initrd_start_page = Page::containing_address(VirtAddr::new(ramdisk_start));
    let initrd_num_pages = usize(ramdisk_len.div_ceil(4096));
    unsafe {
        #[expect(
            clippy::unwrap_used,
            reason = "
            This can only fail if the initrd is not at the address the bootloder says or if it 
            uses huge pages. The first is a violatin of the contract between the bootloader and 
            the kernel, and the second can't happen because the bootloader does not use huge pages.
        "
        )]
        KERNEL_SPACE
            .lock()
            .update_flags(
                initrd_start_page,
                initrd_num_pages,
                PageTableFlags::USER_ACCESSIBLE,
                PageTableFlags::empty(),
            )
            .unwrap();
    }
    initrd
});

#[no_mangle]
extern "C" fn syscall_handler() {
    let regs = unsafe { SYSCALL_REGS };
    let mut retval = 0;
    let mut retval2 = regs.rcx;
    let mut retval3 = regs.rdx;
    let mut retval4 = regs.rsi;
    match regs.rax {
        0 => {
            let rval = if let Some(chr) = char::from_u32(regs.rcx.wrapping_cast()) {
                if chr == '\n' {
                    print!("\r\n");
                } else {
                    print!("{}", chr);
                }
                0
            } else {
                1
            };
            retval = rval;
        }
        1 => TASKING.exit(regs.rcx.wrapping_cast()),
        2 => {
            retval = if regs.rcx == u64::MAX {
                ACTIVE_SPACE
                    .lock()
                    .map_free(usize(regs.rdx), PageTableFlags::from_bits_truncate(regs.rsi))
                    .map_or(0, |x| u64(x.expose_provenance()))
            } else {
                TASKING.current_process(|process| {
                    let mut address_spaces = process.address_spaces().lock();
                    if let Some(space) = address_spaces.get_mut(usize(regs.rcx)) {
                        space
                            .map_free(usize(regs.rdx), PageTableFlags::from_bits_truncate(regs.rsi))
                            .map_or(0, |x| u64(x.expose_provenance()))
                    } else {
                        0
                    }
                })
            };
        }
        3 => {
            retval = u64(INITRD_BUF.as_ptr().expose_provenance());
            retval2 = u64(INITRD_BUF.len());
        }
        4 => {
            #[warn(clippy::expect_used, reason = "FIXME")]
            let address_space = u64(TASKING.current_process(|process| {
                let mut address_spaces = process.address_spaces().lock();
                address_spaces.insert(AddressSpace::new().expect("Failed to create address space"))
            }));
            retval = address_space;
        }
        5 => {
            TASKING
                .current_process(|process| process.address_spaces().lock().remove(usize(regs.rcx)));
        }
        6 => 'call6: {
            let Ok(page) = Page::from_start_address(VirtAddr::new(regs.rdx)) else {
                retval = 1;
                break 'call6;
            };
            let num_pages = usize(regs.rsi);
            let flags = PageTableFlags::from_bits_truncate(regs.rdi);
            let failed = if regs.rcx == u64::MAX {
                ACTIVE_SPACE.lock().map_assert_unused(page, num_pages, flags).is_err()
            } else {
                TASKING.current_process(|process| {
                    let mut address_spaces = process.address_spaces().lock();
                    if let Some(space) = address_spaces.get_mut(usize(regs.rcx)) {
                        space.map_assert_unused(page, num_pages, flags).is_err()
                    } else {
                        true
                    }
                })
            };
            retval = failed.into();
        }
        7 => {
            if let Some(buffer) = get_buffer(regs.rdx) {
                let len = usize(regs.rdi);
                assert!(len <= buffer.len());
                let err = TASKING
                    .current_process(|process| {
                        let mut address_spaces = process.address_spaces().lock();
                        if let Some(space) = address_spaces.get_mut(usize(regs.rcx)) {
                            let buffer_num_pages = buffer.len() / 4096;
                            let buffer_raw = Box::into_raw(buffer);
                            let page = ACTIVE_SPACE
                                .lock()
                                .move_mappings_free(
                                    Page::from_start_address(VirtAddr::new(u64(
                                        buffer_raw.expose_provenance()
                                    )))
                                    .map_err(|_| ())?,
                                    buffer_num_pages,
                                    space,
                                )
                                .map_err(|_| ())?;
                            space.run(|| unsafe {
                                (ptr::with_exposed_provenance_mut::<u8>(usize(regs.rsi)))
                                    .copy_from(page.start_address().as_mut_ptr::<u8>(), len);
                            });
                            space.unmap(page, buffer_num_pages).map_err(|_| ())?;
                            Ok(())
                        } else {
                            Err(())
                        }
                    })
                    .is_err();
                #[expect(clippy::as_conversions, reason = "Needed to convert bool to u64")]
                {
                    retval = err as u64;
                }
            } else {
                retval = 1;
            }
        }
        8 => {
            let space = TASKING
                .current_process(|process| process.address_spaces().lock().remove(usize(regs.rdx)));
            let res = TASKING.new_process(ptr::with_exposed_provenance(usize(regs.rcx)), space);
            if let Ok(pid) = res {
                retval = 0;
                retval2 = u64(pid);
            } else {
                retval = 1;
            }
        }
        9 => {
            #[expect(
                clippy::unwrap_used,
                reason = "Syscalls cannot be called during early boot, the only time when there is no current PID"
            )]
            REGISTERD_PIDS.write().insert(regs.rcx, u64(TASKING.current_pid().unwrap()));
        }
        10 => {
            let id = REGISTERD_PIDS.read().get(&regs.rcx).copied();
            if let Some(id) = id {
                retval = 0;
                retval2 = id;
            } else {
                retval = 1;
            }
        }
        11 => {
            let pid = usize(regs.rcx);
            if let Some(buffer) = get_buffer(regs.rdx) {
                let len = usize(regs.rsi);
                assert!(len <= buffer.len());
                match send_ipc_to(pid, buffer, len, false) {
                    Ok(()) => retval = 0,
                    Err(e) => {
                        println!("ipc_send: Error {e:?}");
                        retval = 1;
                    }
                }
            } else {
                println!("ipc_send: Bad buffer ({})", regs.rdx);
                retval = 1;
            }
        }
        12 => {
            if let Some(msg) =
                TASKING.current_process(|process| process.message_queue().lock().pop())
            {
                #[expect(
                    clippy::unwrap_used,
                    reason = "The message queue only contains valid buffer IDs"
                )]
                let buffer_addr = u64(TASKING
                    .current_process(|process| {
                        *process.data_buffers().lock().get(msg.buffer_idx).unwrap()
                    })
                    .expose_provenance());
                retval2 = u64(msg.len);
                retval = buffer_addr;
                retval3 = u64(msg.buffer_idx);
                retval4 = u64(msg.from);
            } else {
                retval = 0;
            }
        }
        13 => {
            #[expect(
                clippy::unwrap_used,
                reason = "Syscalls cannot be called during early boot, the only time when there is no current PID"
            )]
            let pid = u64(TASKING.current_pid().unwrap());
            retval = pid;
        }
        14 => 'call14: {
            let Ok(page) = Page::from_start_address(VirtAddr::new(regs.rdx)) else {
                retval = 1;
                break 'call14;
            };
            let num_pages = usize(regs.rsi);
            let flags = PageTableFlags::from_bits_truncate(regs.rdi);
            let failed = if regs.rcx == u64::MAX {
                ACTIVE_SPACE.lock().map_only_unused(page, num_pages, flags).is_err()
            } else {
                TASKING.current_process(|process| {
                    let mut address_spaces = process.address_spaces().lock();
                    if let Some(space) = address_spaces.get_mut(usize(regs.rcx)) {
                        space.map_only_unused(page, num_pages, flags).is_err()
                    } else {
                        true
                    }
                })
            };
            retval = failed.into();
        }
        15 => {
            get_buffer(regs.rcx);
        }
        16 => {
            let size = usize(regs.rcx);
            let rounded_size = size.next_multiple_of(4096);
            let mut buffer = Vec::with_capacity_in(rounded_size, &*ACTIVE_SPACE);
            buffer.resize(rounded_size, 0);
            let buffer = buffer.into_boxed_slice();
            let buffer = Box::into_raw(buffer);
            retval = u64(
                TASKING.current_process(|process| process.data_buffers().lock().insert(buffer))
            );
            retval2 = u64(buffer.cast::<u8>().expose_provenance());
            retval3 = u64(rounded_size);
        }
        17 => {
            #[warn(clippy::expect_used, reason = "FIXME")]
            let err = TASKING
                .current_process(|process| {
                    let mut address_spaces = process.address_spaces().lock();
                    let space = address_spaces.get_mut(usize(regs.rcx)).ok_or(())?;
                    let slice_start: *mut u8 = ptr::with_exposed_provenance_mut(usize(regs.rdx));
                    space.run(|| unsafe {
                        slice::from_raw_parts_mut(slice_start, usize(regs.rsi)).fill(0);
                    });
                    Ok::<_, ()>(())
                })
                .is_err();
            #[expect(clippy::as_conversions, reason = "Needed to convert bool to u64")]
            {
                retval = err as u64;
            }
        }
        18 => {
            if TASKING.current_process(|process| process.message_queue().lock().is_empty()) {
                TASKING.sleep(SleepReason::WaitingForIPC);
            }
        }
        19 => (),
        20 => 'call20: {
            let Ok(frame) = PhysFrame::from_start_address(PhysAddr::new(regs.rdx)) else {
                retval = 0;
                break 'call20;
            };
            retval = if regs.rcx == u64::MAX {
                unsafe {
                    ACTIVE_SPACE
                        .lock()
                        .map_free_to(
                            frame,
                            usize(regs.rsi),
                            PageTableFlags::from_bits_truncate(regs.rdi),
                        )
                        .map_or(0, |x| u64(x.expose_provenance()))
                }
            } else {
                TASKING.current_process(|process| {
                    let mut address_spaces = process.address_spaces().lock();
                    if let Some(space) = address_spaces.get_mut(usize(regs.rcx)) {
                        unsafe {
                            space
                                .map_free_to(
                                    frame,
                                    usize(regs.rsi),
                                    PageTableFlags::from_bits_truncate(regs.rdi),
                                )
                                .map_or(0, |x| u64(x.expose_provenance()))
                        }
                    } else {
                        0
                    }
                })
            }
        }
        21 => {
            retval = if regs.rcx == u64::MAX {
                let (start_virt, start_phys) = ACTIVE_SPACE
                    .lock()
                    .map_free_cont_phys(
                        usize(regs.rdx),
                        PageTableFlags::from_bits_truncate(regs.rsi),
                    )
                    .map_or((0, 0), |(ptr, start_phys)| (u64(ptr.expose_provenance()), start_phys));
                retval2 = start_phys;
                start_virt
            } else {
                TASKING.current_process(|process| {
                    let mut address_spaces = process.address_spaces().lock();
                    if let Some(space) = address_spaces.get_mut(usize(regs.rcx)) {
                        let (start_virt, start_phys) = space
                            .map_free_cont_phys(
                                usize(regs.rdx),
                                PageTableFlags::from_bits_truncate(regs.rsi),
                            )
                            .map_or((0, 0), |(ptr, start_phys)| {
                                (u64(ptr.expose_provenance()), start_phys)
                            });
                        retval2 = start_phys;
                        start_virt
                    } else {
                        0
                    }
                })
            };
        }
        22 => 'call22: {
            let Ok(page) = Page::from_start_address(VirtAddr::new(regs.rdx)) else {
                retval = 1;
                break 'call22;
            };
            retval = if regs.rcx == u64::MAX {
                u64::from(ACTIVE_SPACE.lock().unmap(page, usize(regs.rsi)).is_err())
            } else {
                TASKING.current_process(|process| {
                    let mut address_spaces = process.address_spaces().lock();
                    if let Some(space) = address_spaces.get_mut(usize(regs.rcx)) {
                        u64::from(space.unmap(page, usize(regs.rsi)).is_err())
                    } else {
                        1
                    }
                })
            }
        }
        23 => {
            let irq_num = regs.rcx;
            if irq_num < 16 {
                #[expect(
                    clippy::indexing_slicing,
                    reason = "The index has been checked to be in bounds above"
                )]
                IRQ_TASKS.write()[usize(irq_num)] = Some(
                    #[expect(
                        clippy::unwrap_used,
                        reason = "Syscalls cannot be called during early boot, the only time when there is no current PID"
                    )]
                    TASKING.current_pid().unwrap(),
                );
                retval = 0;
            } else {
                retval = 1;
            }
        }
        24 => {
            let pid = usize(regs.rcx);
            if TASKING.wake(pid, 0, SleepReason::NewProcess).is_err() {
                retval = 1;
            } else {
                retval = 0;
            }
        }
        25 => {
            let pid = usize(regs.rcx);
            if TASKING.clear_exited_pid(pid).is_err() {
                retval = 1;
            } else {
                retval = 0;
            }
        }
        26 => {
            let res =
                TASKING.new_thread(ptr::with_exposed_provenance(usize(regs.rcx)), usize(regs.rdx));
            if let Ok(tid) = res {
                retval = 0;
                retval2 = u64(tid);
            } else {
                retval = 1;
            }
        }
        27 => TASKING.exit_thread(),
        28 => {
            TASKING.join_thread(usize(regs.rcx));
        }
        29 => {
            retval = u64(TASKING.new_tls_key());
        }
        30 => TASKING.set_tls_key(usize(regs.rcx), regs.rdx),
        31 => {
            retval = TASKING.get_tls_key(usize(regs.rcx));
        }
        32 => {
            retval = u64(TASKING.new_mutex());
        }
        33 => TASKING.lock_mutex(usize(regs.rcx)),
        34 => {
            retval = u64::from(TASKING.try_lock_mutex(usize(regs.rcx)));
        }
        35 => TASKING.unlock_mutex(usize(regs.rcx)),
        36 => TASKING.drop_mutex(usize(regs.rcx)),
        37 => {
            // Frequency of clock ticks
            retval = 1000;
        }
        38 => {
            retval = NUM_INTERRUPTS.load(Ordering::Relaxed);
        }
        39 => TASKING.time_sleep(regs.rcx),
        40 => {
            retval = u64::from(TASKING.lock_mutex_timeout(usize(regs.rcx), regs.rdx));
        }
        _ => (),
    };
    unsafe { SYSCALL_REGS = regs };
    unsafe {
        asm!(
            "mov rbx, [rip+SYSCALL_REGS+8]",
            "mov rdi, [rip+SYSCALL_REGS+40]",
            "mov rsp, [rip+SYSCALL_REGS+48]",
            "mov rbp, [rip+SYSCALL_REGS+56]",
            "mov r8, [rip+SYSCALL_REGS+64]",
            "mov r9, [rip+SYSCALL_REGS+72]",
            "mov r10, [rip+SYSCALL_REGS+80]",
            "mov r11, [rip+SYSCALL_REGS+88]",
            "mov r12, [rip+SYSCALL_REGS+96]",
            "mov r13, [rip+SYSCALL_REGS+104]",
            "mov r14, [rip+SYSCALL_REGS+112]",
            "mov r15, [rip+SYSCALL_REGS+120]",
            "iretq",
            in("rax") retval, in("rcx") retval2, in("rdx") retval3, in("rsi") retval4, options(noreturn)
        )
    }
}

pub fn register_handler(irq: u8, handler: IrqHandler) -> Result<(), InvalidIrq> {
    *(IRQ_HANDLERS.write().get_mut(usize(irq)).ok_or(InvalidIrq)?) = Some(handler);
    Ok(())
}
