use core::fmt::{self, Display, Formatter};
use serde::{Deserialize, Serialize};

#[repr(i32)]
#[derive(Serialize, Deserialize, Debug, Copy, Clone, PartialEq, Eq)]
pub enum Errno {
    /// Operation not permitted
    EPERM = 1,
    /// No such file or directory
    ENOENT = 2,
    /// No such process
    ESRCH = 3,
    /// Interrupted system call
    EINTR = 4,
    /// Input/output error
    EIO = 5,
    /// No such device or address
    ENXIO = 6,
    /// Argument list too long
    E2BIG = 7,
    /// Exec format error
    ENOEXEC = 8,
    /// Bad file descriptor
    EBADF = 9,
    /// No child processes
    ECHILD = 10,
    /// Resource temporarily unavailable
    EWOULDBLOCK = 11,
    /// Cannot allocate memory
    ENOMEM = 12,
    /// Permission denied
    EACCES = 13,
    /// Bad address
    EFAULT = 14,
    /// Block device required
    ENOTBLK = 15,
    /// Device or resource busy
    EBUSY = 16,
    /// File exists
    EEXIST = 17,
    /// Invalid cross-device link
    EXDEV = 18,
    /// No such device
    ENODEV = 19,
    /// Not a directory
    ENOTDIR = 20,
    /// Is a directory
    EISDIR = 21,
    /// Invalid argument
    EINVAL = 22,
    /// Too many open files in system
    ENFILE = 23,
    /// Too many open files
    EMFILE = 24,
    /// Inappropriate ioctl for device
    ENOTTY = 25,
    /// Text file busy
    ETXTBSY = 26,
    /// File too large
    EFBIG = 27,
    /// No space left on device
    ENOSPC = 28,
    /// Illegal seek
    ESPIPE = 29,
    /// Read-only file system
    EROFS = 30,
    /// Too many links
    EMLINK = 31,
    /// Broken pipe
    EPIPE = 32,
    /// Numerical argument out of domain
    EDOM = 33,
    /// Numerical result out of range
    ERANGE = 34,
    /// Resource deadlock avoided
    EDEADLK = 35,
    /// File name too long
    ENAMETOOLONG = 36,
    /// No locks available
    ENOLCK = 37,
    /// Function not implemented
    ENOSYS = 38,
    /// Directory not empty
    ENOTEMPTY = 39,
    /// Too many levels of symbolic links
    ELOOP = 40,
    /// No message of desired type
    ENOMSG = 42,
    /// Identifier removed
    EIDRM = 43,
    /// Channel number out of range
    ECHRNG = 44,
    /// Level 2 not synchronized
    EL2NSYNC = 45,
    /// Level 3 halted
    EL3HLT = 46,
    /// Level 3 reset
    EL3RST = 47,
    /// Link number out of range
    ELNRNG = 48,
    /// Protocol driver not attached
    EUNATCH = 49,
    /// No CSI structure available
    ENOCSI = 50,
    /// Level 2 halted
    EL2HLT = 51,
    /// Invalid exchange
    EBADE = 52,
    /// Invalid request descriptor
    EBADR = 53,
    /// Exchange full
    EXFULL = 54,
    /// No anode
    ENOANO = 55,
    /// Invalid request code
    EBADRQC = 56,
    /// Invalid slot
    EBADSLT = 57,
    /// Bad font file format
    EBFONT = 59,
    /// Device not a stream
    ENOSTR = 60,
    /// No data available
    ENODATA = 61,
    /// Timer expired
    ETIME = 62,
    /// Out of streams resources
    ENOSR = 63,
    /// Machine is not on the network
    ENONET = 64,
    /// Package not installed
    ENOPKG = 65,
    /// Object is remote
    EREMOTE = 66,
    /// Link has been severed
    ENOLINK = 67,
    /// Advertise error
    EADV = 68,
    /// Srmount error
    ESRMNT = 69,
    /// Communication error on send
    ECOMM = 70,
    /// Protocol error
    EPROTO = 71,
    /// Multihop attempted
    EMULTIHOP = 72,
    /// RFS specific error
    EDOTDOT = 73,
    /// Bad message
    EBADMSG = 74,
    /// Value too large for defined data type
    EOVERFLOW = 75,
    /// Name not unique on network
    ENOTUNIQ = 76,
    /// File descriptor in bad state
    EBADFD = 77,
    /// Remote address changed
    EREMCHG = 78,
    /// Can not access a needed shared library
    ELIBACC = 79,
    /// Accessing a corrupted shared library
    ELIBBAD = 80,
    /// .lib section in a.out corrupted
    ELIBSCN = 81,
    /// Attempting to link in too many shared libraries
    ELIBMAX = 82,
    /// Cannot exec a shared library directly
    ELIBEXEC = 83,
    /// Invalid or incomplete multibyte or wide character
    EILSEQ = 84,
    /// Interrupted system call should be restarted
    ERESTART = 85,
    /// Streams pipe error
    ESTRPIPE = 86,
    /// Too many users
    EUSERS = 87,
    /// Socket operation on non-socket
    ENOTSOCK = 88,
    /// Destination address required
    EDESTADDRREQ = 89,
    /// Message too long
    EMSGSIZE = 90,
    /// Protocol wrong type for socket
    EPROTOTYPE = 91,
    /// Protocol not available
    ENOPROTOOPT = 92,
    /// Protocol not supported
    EPROTONOSUPPORT = 93,
    /// Socket type not supported
    ESOCKTNOSUPPORT = 94,
    /// Operation not supported
    ENOTSUP = 95,
    /// Protocol family not supported
    EPFNOSUPPORT = 96,
    /// Address family not supported by protocol
    EAFNOSUPPORT = 97,
    /// Address already in use
    EADDRINUSE = 98,
    /// Cannot assign requested address
    EADDRNOTAVAIL = 99,
    /// Network is down
    ENETDOWN = 100,
    /// Network is unreachable
    ENETUNREACH = 101,
    /// Network dropped connection on reset
    ENETRESET = 102,
    /// Software caused connection abort
    ECONNABORTED = 103,
    /// Connection reset by peer
    ECONNRESET = 104,
    /// No buffer space available
    ENOBUFS = 105,
    /// Transport endpoint is already connected
    EISCONN = 106,
    /// Transport endpoint is not connected
    ENOTCONN = 107,
    /// Cannot send after transport endpoint shutdown
    ESHUTDOWN = 108,
    /// Too many references: cannot splice
    ETOOMANYREFS = 109,
    /// Connection timed out
    ETIMEDOUT = 110,
    /// Connection refused
    ECONNREFUSED = 111,
    /// Host is down
    EHOSTDOWN = 112,
    /// No route to host
    EHOSTUNREACH = 113,
    /// Operation already in progress
    EALREADY = 114,
    /// Operation now in progress
    EINPROGRESS = 115,
    /// Stale file handle
    ESTALE = 116,
    /// Structure needs cleaning
    EUCLEAN = 117,
    /// Not a XENIX named type file
    ENOTNAM = 118,
    /// No XENIX semaphores available
    ENAVAIL = 119,
    /// Is a named type file
    EISNAM = 120,
    /// Remote I/O error
    EREMOTEIO = 121,
    /// Disk quota exceeded
    EDQUOT = 122,
    /// No medium found
    ENOMEDIUM = 123,
    /// Wrong medium type
    EMEDIUMTYPE = 124,
    /// Operation canceled
    ECANCELED = 125,
    /// Required key not available
    ENOKEY = 126,
    /// Key has expired
    EKEYEXPIRED = 127,
    /// Key has been revoked
    EKEYREVOKED = 128,
    /// Key was rejected by service
    EKEYREJECTED = 129,
    /// Owner died
    EOWNERDEAD = 130,
    /// State not recoverable
    ENOTRECOVERABLE = 131,
    /// Operation not possible due to RF-kill
    ERFKILL = 132,
    /// Memory page has hardware error
    EHWPOISON = 133,
    /// Resource temporarily unavailable
    EAGAIN = 134,
    /// Unexpected EOF
    EEOF = 135,
    /// Invalid data
    EINVALDAT = 136,
    /// Write zero
    EWRZERO = 137,
}

impl TryFrom<i32> for Errno {
    type Error = i32;

    fn try_from(value: i32) -> Result<Self, Self::Error> {
        use Errno::*;
        match value {
            1 => Ok(EPERM),
            2 => Ok(ENOENT),
            3 => Ok(ESRCH),
            4 => Ok(EINTR),
            5 => Ok(EIO),
            6 => Ok(ENXIO),
            7 => Ok(E2BIG),
            8 => Ok(ENOEXEC),
            9 => Ok(EBADF),
            10 => Ok(ECHILD),
            11 => Ok(EWOULDBLOCK),
            12 => Ok(ENOMEM),
            13 => Ok(EACCES),
            14 => Ok(EFAULT),
            15 => Ok(ENOTBLK),
            16 => Ok(EBUSY),
            17 => Ok(EEXIST),
            18 => Ok(EXDEV),
            19 => Ok(ENODEV),
            20 => Ok(ENOTDIR),
            21 => Ok(EISDIR),
            22 => Ok(EINVAL),
            23 => Ok(ENFILE),
            24 => Ok(EMFILE),
            25 => Ok(ENOTTY),
            26 => Ok(ETXTBSY),
            27 => Ok(EFBIG),
            28 => Ok(ENOSPC),
            29 => Ok(ESPIPE),
            30 => Ok(EROFS),
            31 => Ok(EMLINK),
            32 => Ok(EPIPE),
            33 => Ok(EDOM),
            34 => Ok(ERANGE),
            35 => Ok(EDEADLK),
            36 => Ok(ENAMETOOLONG),
            37 => Ok(ENOLCK),
            38 => Ok(ENOSYS),
            39 => Ok(ENOTEMPTY),
            40 => Ok(ELOOP),
            42 => Ok(ENOMSG),
            43 => Ok(EIDRM),
            44 => Ok(ECHRNG),
            45 => Ok(EL2NSYNC),
            46 => Ok(EL3HLT),
            47 => Ok(EL3RST),
            48 => Ok(ELNRNG),
            49 => Ok(EUNATCH),
            50 => Ok(ENOCSI),
            51 => Ok(EL2HLT),
            52 => Ok(EBADE),
            53 => Ok(EBADR),
            54 => Ok(EXFULL),
            55 => Ok(ENOANO),
            56 => Ok(EBADRQC),
            57 => Ok(EBADSLT),
            59 => Ok(EBFONT),
            60 => Ok(ENOSTR),
            61 => Ok(ENODATA),
            62 => Ok(ETIME),
            63 => Ok(ENOSR),
            64 => Ok(ENONET),
            65 => Ok(ENOPKG),
            66 => Ok(EREMOTE),
            67 => Ok(ENOLINK),
            68 => Ok(EADV),
            69 => Ok(ESRMNT),
            70 => Ok(ECOMM),
            71 => Ok(EPROTO),
            72 => Ok(EMULTIHOP),
            73 => Ok(EDOTDOT),
            74 => Ok(EBADMSG),
            75 => Ok(EOVERFLOW),
            76 => Ok(ENOTUNIQ),
            77 => Ok(EBADFD),
            78 => Ok(EREMCHG),
            79 => Ok(ELIBACC),
            80 => Ok(ELIBBAD),
            81 => Ok(ELIBSCN),
            82 => Ok(ELIBMAX),
            83 => Ok(ELIBEXEC),
            84 => Ok(EILSEQ),
            85 => Ok(ERESTART),
            86 => Ok(ESTRPIPE),
            87 => Ok(EUSERS),
            88 => Ok(ENOTSOCK),
            89 => Ok(EDESTADDRREQ),
            90 => Ok(EMSGSIZE),
            91 => Ok(EPROTOTYPE),
            92 => Ok(ENOPROTOOPT),
            93 => Ok(EPROTONOSUPPORT),
            94 => Ok(ESOCKTNOSUPPORT),
            95 => Ok(ENOTSUP),
            96 => Ok(EPFNOSUPPORT),
            97 => Ok(EAFNOSUPPORT),
            98 => Ok(EADDRINUSE),
            99 => Ok(EADDRNOTAVAIL),
            100 => Ok(ENETDOWN),
            101 => Ok(ENETUNREACH),
            102 => Ok(ENETRESET),
            103 => Ok(ECONNABORTED),
            104 => Ok(ECONNRESET),
            105 => Ok(ENOBUFS),
            106 => Ok(EISCONN),
            107 => Ok(ENOTCONN),
            108 => Ok(ESHUTDOWN),
            109 => Ok(ETOOMANYREFS),
            110 => Ok(ETIMEDOUT),
            111 => Ok(ECONNREFUSED),
            112 => Ok(EHOSTDOWN),
            113 => Ok(EHOSTUNREACH),
            114 => Ok(EALREADY),
            115 => Ok(EINPROGRESS),
            116 => Ok(ESTALE),
            117 => Ok(EUCLEAN),
            118 => Ok(ENOTNAM),
            119 => Ok(ENAVAIL),
            120 => Ok(EISNAM),
            121 => Ok(EREMOTEIO),
            122 => Ok(EDQUOT),
            123 => Ok(ENOMEDIUM),
            124 => Ok(EMEDIUMTYPE),
            125 => Ok(ECANCELED),
            126 => Ok(ENOKEY),
            127 => Ok(EKEYEXPIRED),
            128 => Ok(EKEYREVOKED),
            129 => Ok(EKEYREJECTED),
            130 => Ok(EOWNERDEAD),
            131 => Ok(ENOTRECOVERABLE),
            132 => Ok(ERFKILL),
            133 => Ok(EHWPOISON),
            134 => Ok(EAGAIN),
            135 => Ok(EEOF),
            136 => Ok(EINVALDAT),
            137 => Ok(EWRZERO),
            x => Err(x),
        }
    }
}

impl Display for Errno {
    fn fmt(&self, f: &mut Formatter<'_>) -> Result<(), fmt::Error> {
        use Errno::*;
        let errno_name = match self {
            EPERM => "EPERM",
            ENOENT => "ENOENT",
            ESRCH => "ESRCH",
            EINTR => "EINTR",
            EIO => "EIO",
            ENXIO => "ENXIO",
            E2BIG => "E2BIG",
            ENOEXEC => "ENOEXEC",
            EBADF => "EBADF",
            ECHILD => "ECHILD",
            EAGAIN => "EAGAIN",
            EWOULDBLOCK => "EWOULDBLOCK",
            ENOMEM => "ENOMEM",
            EACCES => "EACCES",
            EFAULT => "EFAULT",
            ENOTBLK => "ENOTBLK",
            EBUSY => "EBUSY",
            EEXIST => "EEXIST",
            EXDEV => "EXDEV",
            ENODEV => "ENODEV",
            ENOTDIR => "ENOTDIR",
            EISDIR => "EISDIR",
            EINVAL => "EINVAL",
            ENFILE => "ENFILE",
            EMFILE => "EMFILE",
            ENOTTY => "ENOTTY",
            ETXTBSY => "ETXTBSY",
            EFBIG => "EFBIG",
            ENOSPC => "ENOSPC",
            ESPIPE => "ESPIPE",
            EROFS => "EROFS",
            EMLINK => "EMLINK",
            EPIPE => "EPIPE",
            EDOM => "EDOM",
            ERANGE => "ERANGE",
            EDEADLK => "EDEADLK",
            ENAMETOOLONG => "ENAMETOOLONG",
            ENOLCK => "ENOLCK",
            ENOSYS => "ENOSYS",
            ENOTEMPTY => "ENOTEMPTY",
            ELOOP => "ELOOP",
            ENOMSG => "ENOMSG",
            EIDRM => "EIDRM",
            ECHRNG => "ECHRNG",
            EL2NSYNC => "EL2NSYNC",
            EL3HLT => "EL3HLT",
            EL3RST => "EL3RST",
            ELNRNG => "ELNRNG",
            EUNATCH => "EUNATCH",
            ENOCSI => "ENOCSI",
            EL2HLT => "EL2HLT",
            EBADE => "EBADE",
            EBADR => "EBADR",
            EXFULL => "EXFULL",
            ENOANO => "ENOANO",
            EBADRQC => "EBADRQC",
            EBADSLT => "EBADSLT",
            EBFONT => "EBFONT",
            ENOSTR => "ENOSTR",
            ENODATA => "ENODATA",
            ETIME => "ETIME",
            ENOSR => "ENOSR",
            ENONET => "ENONET",
            ENOPKG => "ENOPKG",
            EREMOTE => "EREMOTE",
            ENOLINK => "ENOLINK",
            EADV => "EADV",
            ESRMNT => "ESRMNT",
            ECOMM => "ECOMM",
            EPROTO => "EPROTO",
            EMULTIHOP => "EMULTIHOP",
            EDOTDOT => "EDOTDOT",
            EBADMSG => "EBADMSG",
            EOVERFLOW => "EOVERFLOW",
            ENOTUNIQ => "ENOTUNIQ",
            EBADFD => "EBADFD",
            EREMCHG => "EREMCHG",
            ELIBACC => "ELIBACC",
            ELIBBAD => "ELIBBAD",
            ELIBSCN => "ELIBSCN",
            ELIBMAX => "ELIBMAX",
            ELIBEXEC => "ELIBEXEC",
            EILSEQ => "EILSEQ",
            ERESTART => "ERESTART",
            ESTRPIPE => "ESTRPIPE",
            EUSERS => "EUSERS",
            ENOTSOCK => "ENOTSOCK",
            EDESTADDRREQ => "EDESTADDRREQ",
            EMSGSIZE => "EMSGSIZE",
            EPROTOTYPE => "EPROTOTYPE",
            ENOPROTOOPT => "ENOPROTOOPT",
            EPROTONOSUPPORT => "EPROTONOSUPPORT",
            ESOCKTNOSUPPORT => "ESOCKTNOSUPPORT",
            ENOTSUP => "ENOTSUP",
            EPFNOSUPPORT => "EPFNOSUPPORT",
            EAFNOSUPPORT => "EAFNOSUPPORT",
            EADDRINUSE => "EADDRINUSE",
            EADDRNOTAVAIL => "EADDRNOTAVAIL",
            ENETDOWN => "ENETDOWN",
            ENETUNREACH => "ENETUNREACH",
            ENETRESET => "ENETRESET",
            ECONNABORTED => "ECONNABORTED",
            ECONNRESET => "ECONNRESET",
            ENOBUFS => "ENOBUFS",
            EISCONN => "EISCONN",
            ENOTCONN => "ENOTCONN",
            ESHUTDOWN => "ESHUTDOWN",
            ETOOMANYREFS => "ETOOMANYREFS",
            ETIMEDOUT => "ETIMEDOUT",
            ECONNREFUSED => "ECONNREFUSED",
            EHOSTDOWN => "EHOSTDOWN",
            EHOSTUNREACH => "EHOSTUNREACH",
            EALREADY => "EALREADY",
            EINPROGRESS => "EINPROGRESS",
            ESTALE => "ESTALE",
            EUCLEAN => "EUCLEAN",
            ENOTNAM => "ENOTNAM",
            ENAVAIL => "ENAVAIL",
            EISNAM => "EISNAM",
            EREMOTEIO => "EREMOTEIO",
            EDQUOT => "EDQUOT",
            ENOMEDIUM => "ENOMEDIUM",
            EMEDIUMTYPE => "EMEDIUMTYPE",
            ECANCELED => "ECANCELED",
            ENOKEY => "ENOKEY",
            EKEYEXPIRED => "EKEYEXPIRED",
            EKEYREVOKED => "EKEYREVOKED",
            EKEYREJECTED => "EKEYREJECTED",
            EOWNERDEAD => "EOWNERDEAD",
            ENOTRECOVERABLE => "ENOTRECOVERABLE",
            ERFKILL => "ERFKILL",
            EHWPOISON => "EHWPOISON",
            EEOF => "EEOF",
            EINVALDAT => "EINVALDAT",
            EWRZERO => "EWRZERO",
        };
        f.write_str(errno_name)
    }
}
