use core::fmt::Display;

use alloc::vec;
use alloc::vec::Vec;
use bitflags::bitflags;
use serde::{Deserialize, Serialize};

use super::{
    errno::Errno,
    message::DissectedMessage,
    rpc_header::{RpcHeader, RpcMessageType},
};

#[derive(Deserialize, Serialize, Clone, Copy, Debug)]
pub enum SeekFrom {
    Start(u64),
    End(i64),
    Current(i64),
}

bitflags! {
    #[derive(Serialize, Deserialize, Clone, Copy, Debug)]
    pub struct PollEvents: u16 {
        const POLLIN = 1<<0 | 1<<1;
        const POLLRDNORM = 1<<0;
        const POLLRDBAND = 1<<1;
        const POLLPRI = 1<<2;
        const POLLOUT = 1<<3;
        const POLLWRNORM = 1<<3;
        const POLLWRBAND = 1<<4;
        const POLLERR = 1<<5;
        const POLLHUP = 1<<6;
        const POLLNVAL = 1<<7;
    }
}

pub enum FileRpcCall {
    Read { fd: u64, len: usize },
    Write { fd: u64 },
    Close { fd: u64 },
    Size { fd: u64 },
    Dup { fd: u64 },
    RegisterPoll { fd: u64, events: PollEvents },
    Poll { poll_id: u64 },
    CancelPoll { poll_id: u64 },
    Seek { fd: u64, pos: SeekFrom },
    Invalid,
}

pub enum FileRpcReturn {
    Read(Result<(), Errno>),
    Write(Result<(), Errno>),
    Close(Result<(), Errno>),
    Size(Result<u64, Errno>),
    Dup(Result<u64, Errno>),
    RegisterPoll(u64),
    Poll(PollEvents),
    CancelPoll,
    Seek(Result<u64, Errno>),
    Invalid,
}

pub enum FileRpcMessage {
    Call(FileRpcCall),
    Return(FileRpcReturn),
}

impl FileRpcMessage {
    pub fn dissect(buffer: &[u8], message: &DissectedMessage<RpcHeader>) -> (Self, Vec<u8>) {
        let rpc_header = message.get_l1();
        match rpc_header.typ {
            RpcMessageType::Call => {
                let mut rem_data = vec![];
                let call = match rpc_header.func {
                    0 => {
                        let (fd, len) = postcard::from_bytes(buffer).unwrap();
                        FileRpcCall::Read { fd, len }
                    }
                    1 => {
                        let (fd, data) = postcard::from_bytes(buffer).unwrap();
                        rem_data = data;
                        FileRpcCall::Write { fd }
                    }
                    2 => {
                        let fd = postcard::from_bytes(buffer).unwrap();
                        FileRpcCall::Close { fd }
                    }
                    3 => {
                        let fd = postcard::from_bytes(buffer).unwrap();
                        FileRpcCall::Size { fd }
                    }
                    4 => {
                        let fd = postcard::from_bytes(buffer).unwrap();
                        FileRpcCall::Dup { fd }
                    }
                    5 => {
                        let (fd, events) = postcard::from_bytes(buffer).unwrap();
                        FileRpcCall::RegisterPoll { fd, events }
                    }
                    6 => {
                        let poll_id = postcard::from_bytes(buffer).unwrap();
                        FileRpcCall::Poll { poll_id }
                    }
                    7 => {
                        let poll_id = postcard::from_bytes(buffer).unwrap();
                        FileRpcCall::CancelPoll { poll_id }
                    }
                    8 => {
                        let (fd, pos) = postcard::from_bytes(buffer).unwrap();
                        FileRpcCall::Seek { fd, pos }
                    }
                    _ => {
                        rem_data = buffer.to_vec();
                        FileRpcCall::Invalid
                    }
                };
                (Self::Call(call), rem_data)
            }
            RpcMessageType::Return => {
                let mut rem_data = vec![];
                let call = match rpc_header.func {
                    0 => {
                        let ret = postcard::from_bytes(buffer).unwrap();
                        match ret {
                            Ok(data) => {
                                rem_data = data;
                                FileRpcReturn::Read(Ok(()))
                            }
                            Err(e) => FileRpcReturn::Read(Err(e)),
                        }
                    }
                    1 => FileRpcReturn::Write(postcard::from_bytes(buffer).unwrap()),
                    2 => FileRpcReturn::Close(postcard::from_bytes(buffer).unwrap()),
                    3 => FileRpcReturn::Size(postcard::from_bytes(buffer).unwrap()),
                    4 => FileRpcReturn::Dup(postcard::from_bytes(buffer).unwrap()),
                    5 => FileRpcReturn::RegisterPoll(postcard::from_bytes(buffer).unwrap()),
                    6 => FileRpcReturn::Poll(postcard::from_bytes(buffer).unwrap()),
                    7 => FileRpcReturn::CancelPoll,
                    8 => FileRpcReturn::Seek(postcard::from_bytes(buffer).unwrap()),
                    _ => {
                        rem_data = buffer.to_vec();
                        FileRpcReturn::Invalid
                    }
                };
                (Self::Return(call), rem_data)
            }
        }
    }
}

impl Display for FileRpcMessage {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        match self {
            Self::Call(call) => match call {
                FileRpcCall::Read { fd, len } => {
                    f.write_fmt(format_args!("Read {len:#x} bytes from FD {fd:#x}"))?;
                }
                FileRpcCall::Write { fd } => f.write_fmt(format_args!("Write to FD {fd:#x}"))?,
                FileRpcCall::Close { fd } => f.write_fmt(format_args!("Close FD {fd:#x}"))?,
                FileRpcCall::Size { fd } => {
                    f.write_fmt(format_args!("Get size of FD {fd:#x}"))?;
                }
                FileRpcCall::Dup { fd } => f.write_fmt(format_args!("Duplicate FD {fd:#x}"))?,
                FileRpcCall::RegisterPoll { fd, events } => f.write_fmt(format_args!(
                    "Register a poll for FD {fd:#x} for events {events:?}"
                ))?,
                FileRpcCall::Poll { poll_id } => {
                    f.write_fmt(format_args!("Poll on poll ID {poll_id}"))?;
                }
                FileRpcCall::CancelPoll { poll_id } => {
                    f.write_fmt(format_args!("Cancel poll with ID {poll_id}"))?;
                }
                FileRpcCall::Seek { fd, pos } => {
                    f.write_fmt(format_args!("Seek FD {fd:#x} to position {pos:?}"))?;
                }
                FileRpcCall::Invalid => f.write_fmt(format_args!("Invalid call!"))?,
            },
            Self::Return(ret) => match ret {
                FileRpcReturn::Read(ret) => f.write_fmt(format_args!("Read return {ret:?}"))?,
                FileRpcReturn::Write(ret) => f.write_fmt(format_args!("Write return {ret:?}"))?,
                FileRpcReturn::Close(ret) => f.write_fmt(format_args!("Close return {ret:?}"))?,
                FileRpcReturn::Size(ret) => f.write_fmt(format_args!("Size return {ret:?}"))?,
                FileRpcReturn::Dup(ret) => f.write_fmt(format_args!("Duplicate return {ret:?}"))?,
                FileRpcReturn::RegisterPoll(ret) => {
                    f.write_fmt(format_args!("Poll register return {ret}"))?;
                }
                FileRpcReturn::Poll(ret) => f.write_fmt(format_args!("Poll return {ret:?}"))?,
                FileRpcReturn::CancelPoll => f.write_fmt(format_args!("Cancel poll return"))?,
                FileRpcReturn::Seek(ret) => f.write_fmt(format_args!("Seek return {ret:?}"))?,
                FileRpcReturn::Invalid => f.write_fmt(format_args!("Invalid return!"))?,
            },
        };
        Ok(())
    }
}
