use alloc::fmt::Display;
use cast::{u32, usize};

pub struct HexdumpLayer<'a> {
    data: &'a [u8],
    max_len: usize,
}

impl<'a> HexdumpLayer<'a> {
    pub fn new(data: &'a [u8], max_len: usize) -> Self {
        Self { data, max_len }
    }
}

impl Display for HexdumpLayer<'_> {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        let hexdump_len = usize::min(self.data.len(), self.max_len);
        let data = &self.data[0..hexdump_len];
        let num_groups = data.len().div_ceil(16);
        let offset_char_width = if num_groups == 0 {
            1
        } else if num_groups == 1 {
            2
        } else {
            usize(num_groups.next_power_of_two().ilog2().div_ceil(4) + 1)
        };
        let mut chunk_iter = data.chunks_exact(16);
        for (group_num, chunk) in (&mut chunk_iter).enumerate() {
            if group_num == 0 {
                f.write_fmt(format_args!("{:01$}:", 0, offset_char_width))?;
            } else {
                f.write_fmt(format_args!("\n{:01$}0:", group_num, offset_char_width - 1))?;
            }
            for &byte in chunk {
                f.write_fmt(format_args!(" {byte:02x}"))?;
            }
            f.write_str(" |")?;
            for &byte in chunk {
                let ch = char::from_u32(u32(byte)).unwrap_or('.');
                let ch = if ch.is_ascii() && !ch.is_ascii_control() { ch } else { '.' };
                f.write_fmt(format_args!("{ch}"))?;
            }
        }
        let chunk = chunk_iter.remainder();
        if !chunk.is_empty() {
            let group_num = num_groups - 1;
            if group_num == 0 {
                f.write_fmt(format_args!("{:01$}:", 0, offset_char_width))?;
            } else {
                f.write_fmt(format_args!("\n{:01$}0:", group_num, offset_char_width - 1))?;
            }
            for &byte in chunk {
                f.write_fmt(format_args!(" {byte:02x}"))?;
            }
            let missing_bytes = 16 - chunk.len();
            for _ in 0..missing_bytes {
                f.write_str("   ")?;
            }
            f.write_str(" |")?;
            for &byte in chunk {
                let ch = char::from_u32(u32(byte)).unwrap_or('.');
                let ch = if ch.is_ascii() && !ch.is_ascii_control() { ch } else { '.' };
                f.write_fmt(format_args!("{ch}"))?;
            }
        }
        Ok(())
    }
}
