use alloc::fmt::{self, Display};
use alloc::string::{String, ToString};

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum IpcFrom {
    Kernel,
    Process(usize),
}

impl IpcFrom {
    pub fn as_pid(self) -> usize {
        match self {
            Self::Kernel => usize::MAX,
            Self::Process(pid) => pid,
        }
    }
}

impl Display for IpcFrom {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::Kernel => f.write_str("Kernel"),
            Self::Process(pid) => f.write_fmt(format_args!("{pid}")),
        }
    }
}

pub trait MessageHeader: Display {}
impl<T: Display> MessageHeader for T {}

pub struct DissectedMessage<L1 = NoHeader, L2 = NoHeader, L3 = NoHeader> {
    pub ipc_from: IpcFrom,
    pub ipc_dest: usize,
    pub ipc_proto: u16,
    l1: L1,
    l2: L2,
    l3: L3,
}

pub struct NoHeader;

impl DissectedMessage<NoHeader, NoHeader, NoHeader> {
    pub fn new(ipc_from: IpcFrom, ipc_dest: usize, ipc_proto: u16) -> Self {
        Self { ipc_from, ipc_dest, ipc_proto, l1: NoHeader, l2: NoHeader, l3: NoHeader }
    }

    pub fn set_l1<T: MessageHeader>(self, l1: T) -> DissectedMessage<T, NoHeader, NoHeader> {
        DissectedMessage {
            ipc_from: self.ipc_from,
            ipc_dest: self.ipc_dest,
            ipc_proto: self.ipc_proto,
            l1,
            l2: NoHeader,
            l3: NoHeader,
        }
    }
}

impl<L1: MessageHeader> DissectedMessage<L1, NoHeader, NoHeader> {
    pub fn get_l1(&self) -> &L1 {
        &self.l1
    }

    pub fn set_l2<T: MessageHeader>(self, l2: T) -> DissectedMessage<L1, T, NoHeader> {
        DissectedMessage {
            ipc_from: self.ipc_from,
            ipc_dest: self.ipc_dest,
            ipc_proto: self.ipc_proto,
            l1: self.l1,
            l2,
            l3: NoHeader,
        }
    }
}

impl<L1: MessageHeader, L2: MessageHeader> DissectedMessage<L1, L2, NoHeader> {
    pub fn get_l2(&self) -> &L2 {
        &self.l2
    }

    pub fn set_l3<T: MessageHeader>(self, l3: T) -> DissectedMessage<L1, L2, T> {
        DissectedMessage {
            ipc_from: self.ipc_from,
            ipc_dest: self.ipc_dest,
            ipc_proto: self.ipc_proto,
            l1: self.l1,
            l2: self.l2,
            l3,
        }
    }
}

impl<L1: MessageHeader, L2: MessageHeader, L3: MessageHeader> DissectedMessage<L1, L2, L3> {
    pub fn get_l3(&self) -> &L3 {
        &self.l3
    }
}

impl<L1, L2, L3> DissectedMessage<L1, L2, L3> {
    fn fmt_ipc_header(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.write_fmt(format_args!(
            "IPC: {} sending to {}, protocol {}",
            self.ipc_from, self.ipc_dest, self.ipc_proto,
        ))
    }
}

impl<L1: MessageHeader, L2, L3> DissectedMessage<L1, L2, L3> {
    fn fmt_l1(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let l1_raw = self.l1.to_string();
        let l1_indented =
            l1_raw.split('\n').map(|line| "    ".to_string() + line + "\n").collect::<String>();
        f.write_str(l1_indented.strip_suffix("\n").unwrap())
    }
}

impl<L1: MessageHeader, L2: MessageHeader, L3> DissectedMessage<L1, L2, L3> {
    fn fmt_l2(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let l2_raw = self.l2.to_string();
        let l2_indented =
            l2_raw.split('\n').map(|line| "        ".to_string() + line + "\n").collect::<String>();
        f.write_str(l2_indented.strip_suffix("\n").unwrap())
    }
}

impl<L1: MessageHeader, L2: MessageHeader, L3: MessageHeader> DissectedMessage<L1, L2, L3> {
    fn fmt_l3(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let l3_raw = self.l3.to_string();
        let l3_indented = l3_raw
            .split('\n')
            .map(|line| "            ".to_string() + line + "\n")
            .collect::<String>();
        f.write_str(l3_indented.strip_suffix("\n").unwrap())
    }
}

impl Display for DissectedMessage<NoHeader, NoHeader, NoHeader> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.write_str("-----------------------------------------\n")?;
        self.fmt_ipc_header(f)?;
        f.write_str("\n")?;
        f.write_str("-----------------------------------------\n")?;
        Ok(())
    }
}

impl<L1: MessageHeader> Display for DissectedMessage<L1, NoHeader, NoHeader> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.write_str("-----------------------------------------\n")?;
        self.fmt_ipc_header(f)?;
        f.write_str("\n")?;
        self.fmt_l1(f)?;
        f.write_str("\n")?;
        f.write_str("-----------------------------------------\n")?;
        Ok(())
    }
}

impl<L1: MessageHeader, L2: MessageHeader> Display for DissectedMessage<L1, L2, NoHeader> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.write_str("-----------------------------------------\n")?;
        self.fmt_ipc_header(f)?;
        f.write_str("\n")?;
        self.fmt_l1(f)?;
        f.write_str("\n")?;
        self.fmt_l2(f)?;
        f.write_str("\n")?;
        f.write_str("-----------------------------------------\n")?;
        Ok(())
    }
}

impl<L1: MessageHeader, L2: MessageHeader, L3: MessageHeader> Display
    for DissectedMessage<L1, L2, L3>
{
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.write_str("-----------------------------------------\n")?;
        self.fmt_ipc_header(f)?;
        f.write_str("\n")?;
        self.fmt_l1(f)?;
        f.write_str("\n")?;
        self.fmt_l2(f)?;
        f.write_str("\n")?;
        self.fmt_l3(f)?;
        f.write_str("\n")?;
        f.write_str("-----------------------------------------\n")?;

        Ok(())
    }
}
