use alloc::fmt::Display;

use hashbrown::HashMap;
use spin::{Lazy, Mutex};

use super::message::DissectedMessage;

static RPC_CALL_MAP: Lazy<Mutex<HashMap<(usize, u64), (u16, u16)>>> =
    Lazy::new(|| Mutex::new(HashMap::new()));

#[derive(Debug, Copy, Clone, PartialEq, Eq)]
pub enum RpcMessageType {
    Call,
    Return,
}

pub struct RpcHeader {
    pub typ: RpcMessageType,
    pub call_id: u64,
    pub proto: u16,
    pub func: u16,
    //from: IpcFrom,
    //dest: usize,
}

impl RpcHeader {
    pub fn dissect<'a>(buffer: &'a [u8], message: &DissectedMessage) -> (Self, &'a [u8]) {
        let typ = match buffer[0] {
            0 => RpcMessageType::Call,
            1 => RpcMessageType::Return,
            _ => panic!(),
        };
        let call_id = u64::from_ne_bytes(buffer[1..9].try_into().unwrap());
        let (proto, func) = if typ == RpcMessageType::Call {
            let proto = u16::from_ne_bytes(buffer[9..11].try_into().unwrap());
            let func = u16::from_ne_bytes(buffer[11..13].try_into().unwrap());
            RPC_CALL_MAP.lock().insert((message.ipc_from.as_pid(), call_id), (proto, func));
            (proto, func)
        } else if let Some((proto, func)) = RPC_CALL_MAP.lock().remove(&(message.ipc_dest, call_id))
        {
            (proto, func)
        } else {
            (u16::MAX, u16::MAX)
        };
        let slf = Self {
            typ,
            call_id,
            proto,
            func,
            //from: message.ipc_from,
            //dest: message.ipc_dest,
        };
        (slf, &buffer[13..])
    }
}

impl Display for RpcHeader {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        match self.typ {
            RpcMessageType::Call => {
                f.write_fmt(format_args!(
                    "RPC: call w/ID {}, protocol {}, function {}",
                    self.call_id, self.proto, self.func
                ))?;
            }
            RpcMessageType::Return => {
                if self.proto == u16::MAX && self.func == u16::MAX {
                    f.write_fmt(format_args!("RPC: return for call {} (Not seen!)", self.call_id))?;
                } else {
                    f.write_fmt(format_args!(
                        "RPC: return for call {} (proto {}, func {})",
                        self.call_id, self.proto, self.func
                    ))?;
                }
            }
        }
        Ok(())
    }
}
