use crate::{
    gdt,
    interrupts::{send_ipc_to, REGISTERD_PIDS},
    pit, println,
    virtual_memory::{ASpaceMutex, AddressSpace, PagingError, ACTIVE_SPACE, KERNEL_SPACE},
};
use alloc::{boxed::Box, collections::VecDeque, vec, vec::Vec};
use cast::{u64, usize};
use core::{
    arch::{asm, naked_asm},
    ptr::{addr_of, addr_of_mut},
    sync::atomic::{AtomicBool, Ordering},
};
use crossbeam_queue::SegQueue;
use slab::Slab;
use spin::{Lazy, Mutex, Once, RwLock};
use x86_64::{
    registers::control::{Cr0, Cr0Flags},
    structures::paging::{Page, PageTableFlags},
    VirtAddr,
};

#[naked]
extern "C" fn switch_to_asm(current_stack: *mut *mut usize, next_stack: *mut usize) {
    unsafe {
        naked_asm!(
            "push rbp",
            "push rbx",
            "push r12",
            "push r13",
            "push r14",
            "push r15",
            "mov [rdi], rsp",
            "mov rsp, rsi",
            "pop r15",
            "pop r14",
            "pop r13",
            "pop r12",
            "pop rbx",
            "pop rbp",
            "ret",
        );
    }
}

#[naked]
extern "C" fn switch_to_asm_exit(next_stack: *mut usize) {
    unsafe {
        naked_asm!(
            "mov rsp, rdi",
            "pop r15",
            "pop r14",
            "pop r13",
            "pop r12",
            "pop rbx",
            "pop rbp",
            "ret",
        );
    }
}

#[naked]
extern "C" fn task_init() {
    unsafe {
        naked_asm!(
            "pop rcx",  // Get the user stack pointer
            "pop rbx",  // Get the entry point
            "pop rdi",  // Get argument to the entry point
            "push 43",  // Push the stack segment selector - same as data
            "push rcx", // Push the stack pointer
            "pushfq",   // Get the flags into RAX
            "pop rax",
            "or rax, 0x200", // Enable interrupts in the stored copy
            "push rax",      // Push the flags
            "push 51",       // Push the code selector
            "push rbx",      // Push the entry point
            "iretq",         // Return from the fake interrupt and enter user mode
        )
    }
}

#[derive(Copy, Clone, Debug, PartialEq, Eq)]
pub enum SleepReason {
    WaitingForIPC,
    NewProcess,
    Exited,
    JoinThread(usize),
    LockedMutex(usize),
    TimeSleep,
}

static INITIAL_FPU_STATE: Once<FpuState> = Once::new();

pub fn store_initial_fpu_state() {
    let mut state = FpuState::new();
    state.save();
    INITIAL_FPU_STATE.call_once(|| state);
}

#[repr(align(16))]
#[derive(Debug, Clone, Copy)]
struct FpuState {
    state: [u8; 512],
}

impl FpuState {
    fn new() -> Self {
        Self { state: [0; 512] }
    }

    fn save(&mut self) {
        unsafe { asm!("fxsave [rax]", in("rax") &mut self.state) };
    }

    fn load(&self) {
        unsafe { asm!("fxrstor [rax]", in("rax") &self.state) };
    }
}

#[derive(Debug)]
pub struct Process {
    address_space: Option<AddressSpace>,
    address_spaces: Mutex<Slab<AddressSpace>>,
    data_buffers: Mutex<Slab<*mut [u8]>>,
    message_queue: Mutex<SegQueue<IpcMessage>>,
    threads: RwLock<Slab<Thread>>,
    main_thread: usize,
    num_threads: usize,
    num_tls: usize,
    mutexes: Mutex<Slab<bool>>,
}

impl Process {
    fn new(
        address_space: AddressSpace,
        entry_point: *const extern "C" fn(usize) -> !,
    ) -> Result<Self, PagingError> {
        let mut proc = Self {
            address_space: Some(address_space),
            address_spaces: Mutex::new(Slab::new()),
            data_buffers: Mutex::new(Slab::new()),
            message_queue: Mutex::new(SegQueue::new()),
            threads: RwLock::new(Slab::new()),
            main_thread: 0,
            num_threads: 0,
            num_tls: 0,
            mutexes: Mutex::new(Slab::new()),
        };
        let main_thread = proc.new_thread(entry_point, 0)?;
        proc.main_thread = main_thread;
        Ok(proc)
    }

    fn new_thread(
        &mut self,
        entry_point: *const extern "C" fn(usize) -> !,
        argument: usize,
    ) -> Result<usize, PagingError> {
        let stack_start = 0xFFF_FF80_0000 - (u64(self.num_threads) * (16 * 4096));
        let mut kernel_stack = Vec::with_capacity_in(KSTACK_SIZE, &*KERNEL_SPACE);

        kernel_stack.resize(KSTACK_SIZE - 0x4, 0);
        #[expect(clippy::as_conversions, reason = "Needed to get address of function")]
        kernel_stack.push(task_init as usize);
        kernel_stack.push(usize(stack_start) + (16 * 4096));
        kernel_stack.push(entry_point.expose_provenance());
        kernel_stack.push(argument);
        let mut kernel_stack = kernel_stack.into_boxed_slice();
        if let Some(address_space) = self.address_space.as_mut() {
            address_space.map_assert_unused(
                #[expect(
                    clippy::unwrap_used,
                    reason = "from_start_address requires the address to be page aligned, which it is."
                )]
                Page::from_start_address(VirtAddr::new(stack_start)).unwrap(),
                16,
                PageTableFlags::USER_ACCESSIBLE,
            )?;
        } else {
            ACTIVE_SPACE.lock().map_assert_unused(
                #[expect(
                    clippy::unwrap_used,
                    reason = "from_start_address requires the address to be page aligned, which it is."
                )]
                Page::from_start_address(VirtAddr::new(stack_start)).unwrap(),
                16,
                PageTableFlags::USER_ACCESSIBLE,
            )?;
        }
        let sleeping =
            if self.threads.read().is_empty() { Some(SleepReason::NewProcess) } else { None };
        let thread = Thread {
            #[expect(
                clippy::indexing_slicing,
                reason = "Stack length is 0x1_0000, this cannot panic"
            )]
            kernel_esp: &mut kernel_stack[KSTACK_SIZE - 10],
            #[expect(
                clippy::indexing_slicing,
                reason = "Stack length is 0x1_0000, this cannot panic"
            )]
            kernel_esp_top: VirtAddr::from_ptr(
                addr_of!(kernel_stack[KSTACK_SIZE - 1]).wrapping_add(1),
            ),
            kernel_stack,
            sleeping: RwLock::new(sleeping),
            tls_values: Mutex::new(vec![0; self.num_tls]),
            fpu_state: Mutex::new(*INITIAL_FPU_STATE.get().unwrap()),
        };
        let idx = self.threads.write().insert(thread);
        self.num_threads += 1;
        Ok(idx)
    }

    pub fn threads(&self) -> &RwLock<Slab<Thread>> {
        &self.threads
    }
}

#[derive(Debug)]
pub struct Thread {
    sleeping: RwLock<Option<SleepReason>>,
    kernel_stack: Box<[usize], &'static ASpaceMutex>,
    kernel_esp: *mut usize,
    kernel_esp_top: VirtAddr,
    tls_values: Mutex<Vec<u64>>,
    fpu_state: Mutex<FpuState>,
}

pub struct IpcMessage {
    pub buffer_idx: usize,
    pub len: usize,
    pub from: usize,
}

impl Process {
    pub fn address_spaces(&self) -> &Mutex<Slab<AddressSpace>> {
        &self.address_spaces
    }

    pub fn data_buffers(&self) -> &Mutex<Slab<*mut [u8]>> {
        &self.data_buffers
    }

    pub fn message_queue(&self) -> &Mutex<SegQueue<IpcMessage>> {
        &self.message_queue
    }

    pub fn address_space_mut(&mut self) -> Option<&mut AddressSpace> {
        self.address_space.as_mut()
    }
}

unsafe impl Send for Process {}
unsafe impl Sync for Process {}

#[derive(Copy, Clone, Debug)]
pub struct InvalidPid;

pub static TASKING: Lazy<Tasking> = Lazy::new(|| Tasking {
    processes: RwLock::new(Slab::new()),
    ready_to_run: Mutex::new(VecDeque::new()),
    current_pid: RwLock::new(None),
    current_tid: RwLock::new(None),
    freeable_kstacks: Mutex::new(Vec::new()),
    wfi_loop: AtomicBool::new(false),
    current_fpu_context: Mutex::new(None),
});

#[derive(Debug)]
pub struct Tasking {
    processes: RwLock<Slab<Process>>,
    ready_to_run: Mutex<VecDeque<(usize, usize)>>,
    current_pid: RwLock<Option<usize>>,
    current_tid: RwLock<Option<usize>>,
    freeable_kstacks: Mutex<Vec<Box<[usize], &'static ASpaceMutex>>>,
    wfi_loop: AtomicBool,
    current_fpu_context: Mutex<Option<(usize, usize)>>,
}

pub const KSTACK_SIZE: usize = (4 * 4096) / 8;

impl Tasking {
    pub fn new_process(
        &self,
        entry_point: *const extern "C" fn(usize) -> !,
        address_space: AddressSpace,
    ) -> Result<usize, PagingError> {
        let process = Process::new(address_space, entry_point)?;
        let pid = self.processes.write().insert(process);
        if let Some(&proc_man_pid) = REGISTERD_PIDS.read().get(&3) {
            let mut varint_buf = unsigned_varint::encode::u64_buffer();
            let mut buffer = Vec::new_in(&*ACTIVE_SPACE);
            buffer.extend_from_slice(&0u16.to_le_bytes());
            buffer.push(0);
            buffer.extend_from_slice(&0u64.to_le_bytes());
            buffer.extend_from_slice(&8u16.to_le_bytes());
            buffer.extend_from_slice(&6u16.to_le_bytes());
            buffer.extend_from_slice(unsigned_varint::encode::u64(u64(pid), &mut varint_buf));
            if let Some(current_pid) = *(self.current_pid.read()) {
                buffer.push(1);
                buffer.extend_from_slice(unsigned_varint::encode::u64(
                    u64(current_pid),
                    &mut varint_buf,
                ));
            } else {
                buffer.push(0);
            }
            let len = buffer.len();
            buffer.resize(len.next_multiple_of(4096), 0);
            let buffer = buffer.into_boxed_slice();
            #[expect(
                clippy::expect_used,
                reason = "The tasking code in the kernel and proc_man CANNOT lose sync. Failure to communicate is fatal."
            )]
            send_ipc_to(usize(proc_man_pid), buffer, len, true)
                .expect("Failed to send exit message to proc_man");
        } else {
            println!("[TASKING] No process manager when creating PID {pid}");
        }
        Ok(pid)
    }

    pub fn new_thread(
        &self,
        entry_point: *const extern "C" fn(usize) -> !,
        argument: usize,
    ) -> Result<usize, PagingError> {
        let current_pid = self.current_pid().unwrap();
        let new_tid =
            self.current_process_mut(|process| process.new_thread(entry_point, argument))?;
        self.ready_to_run.lock().push_back((current_pid, new_tid));
        Ok(new_tid)
    }

    pub fn ok_to_yield(&self) -> bool {
        !(self.freeable_kstacks.is_locked()
            || (self.current_pid.reader_count() > 0)
            || (self.current_pid.writer_count() > 0)
            || (self.current_tid.reader_count() > 0)
            || (self.current_tid.writer_count() > 0)
            || self.ready_to_run.is_locked()
            || (self.processes.reader_count() > 0)
            || (self.processes.writer_count() > 0)
            || KERNEL_SPACE.is_locked()
            || self.wfi_loop.load(Ordering::Relaxed))
    }

    pub fn task_yield(&self) {
        loop {
            self.freeable_kstacks.lock().clear();
            let Some(current_pid) = self.current_pid() else {
                self.wfi_loop.store(false, Ordering::Relaxed);
                break;
            };
            let rtr_head = self.ready_to_run.lock().pop_front();
            if let Some((next_process_pid, next_process_tid)) = rtr_head {
                self.wfi_loop.store(false, Ordering::Relaxed);
                if Some(next_process_pid) == self.current_pid()
                    && Some(next_process_tid) == self.current_tid()
                {
                    println!("Yielding to current thread! Returning");
                    break;
                }
                #[warn(clippy::unwrap_used, reason = "FIXME(?)")]
                if current_pid != next_process_pid {
                    let current_address_space = self
                    .process_mut(next_process_pid, |process| {
                        #[expect(
                            clippy::expect_used,
                            reason = "This expect checks a critical invariant. If this fails, the kernel MUST panic"
                        )]
                        process
                            .address_space
                            .take()
                            .expect("Non-current process has active page table")
                            .activate()
                    })
                    .unwrap();
                    self.current_process_mut(|process| {
                        process.address_space = Some(current_address_space);
                    });
                }
                if self.current_thread(|thread| thread.sleeping.read().is_none()) {
                    self.ready_to_run.lock().push_back((current_pid, self.current_tid().unwrap()));
                }
                let curr_stack = self.current_thread_mut(|thread| addr_of_mut!(thread.kernel_esp));
                *self.current_pid.write() = Some(next_process_pid);
                *self.current_tid.write() = Some(next_process_tid);
                let kernel_esp = self.current_thread(|thread| {
                    gdt::set_tss_stack(thread.kernel_esp_top);
                    thread.kernel_esp
                });
                unsafe {
                    Cr0::update(|cr0| {
                        cr0.insert(Cr0Flags::TASK_SWITCHED);
                    });
                }
                switch_to_asm(curr_stack, kernel_esp);
                break;
            } else if self.current_thread(|thread| thread.sleeping.read().is_some()) {
                self.wfi_loop.store(true, Ordering::Relaxed);
                x86_64::instructions::interrupts::enable_and_hlt();
                x86_64::instructions::interrupts::disable();
            } else {
                self.wfi_loop.store(false, Ordering::Relaxed);
                break;
            }
        }
    }

    pub fn current_pid(&self) -> Option<usize> {
        *self.current_pid.read()
    }

    pub fn current_tid(&self) -> Option<usize> {
        *self.current_tid.read()
    }

    pub fn exit(&self, code: u8) -> ! {
        if let Some(current_pid) = self.current_pid() {
            let current_fpu_context = *self.current_fpu_context.lock();
            if let Some((fpu_pid, _fpu_tid)) = current_fpu_context {
                if current_pid == fpu_pid {
                    *self.current_fpu_context.lock() = None;
                }
            }
            self.ready_to_run.lock().retain(|&(ent_pid, _ent_tid)| ent_pid != current_pid);
            if let Some(&proc_man_pid) = REGISTERD_PIDS.read().get(&3) {
                let mut varint_buf = unsigned_varint::encode::u64_buffer();
                let mut buffer = Vec::new_in(&*ACTIVE_SPACE);
                buffer.extend_from_slice(&0u16.to_le_bytes());
                buffer.push(0);
                buffer.extend_from_slice(&0u64.to_le_bytes());
                buffer.extend_from_slice(&8u16.to_le_bytes());
                buffer.extend_from_slice(&4u16.to_le_bytes());
                buffer.extend_from_slice(unsigned_varint::encode::u64(
                    u64(current_pid),
                    &mut varint_buf,
                ));
                buffer.push(code);
                let len = buffer.len();
                buffer.resize(len.next_multiple_of(4096), 0);
                let buffer = buffer.into_boxed_slice();
                #[expect(
                    clippy::expect_used,
                    reason = "The tasking code in the kernel and proc_man CANNOT lose sync. Failure to communicate is fatal."
                )]
                send_ipc_to(usize(proc_man_pid), buffer, len, true)
                    .expect("Failed to send exit message to proc_man");
            } else {
                println!(
                    "[TASKING] No process manager when PID {} exited with code {code}",
                    current_pid
                );
            }
        }
        loop {
            let rtr_head = self.ready_to_run.lock().pop_front();
            if let Some((next_process_pid, next_process_tid)) = rtr_head {
                self.wfi_loop.store(false, Ordering::Relaxed);
                #[warn(clippy::indexing_slicing, reason = "FIXME(?)")]
                if self.current_pid.read().is_some() {
                    self.current_process(|process| {
                        for (_, thread) in &*process.threads().read() {
                            *thread.sleeping.write() = Some(SleepReason::Exited);
                        }
                    });
                }
                *self.current_pid.write() = Some(next_process_pid);
                *self.current_tid.write() = Some(next_process_tid);
                self
                    .current_process_mut(|process| {
                        #[expect(
                            clippy::expect_used,
                            reason = "This expect checks a critical invariant. If this fails, the kernel MUST panic"
                        )]
                        process
                            .address_space
                            .take()
                            .expect("Non-current process has active page table")
                            .activate();
                    });
                let kernel_esp = self.current_thread_mut(|thread| {
                    gdt::set_tss_stack(thread.kernel_esp_top);
                    thread.kernel_esp
                });
                unsafe {
                    Cr0::update(|cr0| {
                        cr0.insert(Cr0Flags::TASK_SWITCHED);
                    });
                }
                switch_to_asm_exit(kernel_esp);
                unreachable!()
            } else {
                self.wfi_loop.store(true, Ordering::Relaxed);
                x86_64::instructions::interrupts::enable_and_hlt();
                x86_64::instructions::interrupts::disable();
            }
        }
    }

    pub fn exit_thread(&self) -> ! {
        loop {
            #[warn(clippy::unwrap_used, reason = "FIXME(?)")]
            let current_pid = self.current_pid().unwrap();
            let current_fpu_context = *self.current_fpu_context.lock();
            if let Some((fpu_pid, fpu_tid)) = current_fpu_context {
                if current_pid == fpu_pid && self.current_tid().unwrap() == fpu_tid {
                    *self.current_fpu_context.lock() = None;
                }
            }
            let rtr_head = self.ready_to_run.lock().pop_front();
            if let Some((next_process_pid, next_process_tid)) = rtr_head {
                self.wfi_loop.store(false, Ordering::Relaxed);
                #[warn(clippy::indexing_slicing, reason = "FIXME(?)")]
                self.current_thread(|thread| {
                    *thread.sleeping.write() = Some(SleepReason::Exited);
                });
                self.current_process(|process| {
                    for (tid, _) in &*process.threads().read() {
                        self.wake(
                            current_pid,
                            tid,
                            SleepReason::JoinThread(self.current_tid().unwrap()),
                        )
                        .unwrap();
                    }
                });
                #[warn(clippy::unwrap_used, reason = "FIXME(?)")]
                if current_pid != next_process_pid {
                    let current_address_space = self
                    .process_mut(next_process_pid, |process| {
                        #[expect(
                            clippy::expect_used,
                            reason = "This expect checks a critical invariant. If this fails, the kernel MUST panic"
                        )]
                        process
                            .address_space
                            .take()
                            .expect("Non-current process has active page table")
                            .activate()
                    })
                    .unwrap();
                    self.current_process_mut(|process| {
                        process.address_space = Some(current_address_space);
                    });
                }
                *self.current_pid.write() = Some(next_process_pid);
                *self.current_tid.write() = Some(next_process_tid);
                let kernel_esp = self.current_thread_mut(|thread| {
                    gdt::set_tss_stack(thread.kernel_esp_top);
                    thread.kernel_esp
                });
                unsafe {
                    Cr0::update(|cr0| {
                        cr0.insert(Cr0Flags::TASK_SWITCHED);
                    });
                }
                switch_to_asm_exit(kernel_esp);
                unreachable!()
            } else {
                self.wfi_loop.store(true, Ordering::Relaxed);
                x86_64::instructions::interrupts::enable_and_hlt();
                x86_64::instructions::interrupts::disable();
            }
        }
    }

    pub fn join_thread(&self, tid: usize) {
        if self
            .curr_proc_thread(tid, |thread| *thread.sleeping.read() == Some(SleepReason::Exited))
            .unwrap()
        {
            return;
        }
        self.sleep(SleepReason::JoinThread(tid));
    }

    pub fn clear_exited_pid(&self, pid: usize) -> Result<(), ()> {
        if self.current_pid() == Some(pid) {
            return Err(());
        }
        let mut processes = self.processes.write();
        let process = processes.remove(pid);
        self.ready_to_run.lock().retain(|&(ent_pid, _ent_tid)| ent_pid != pid);
        let threads = process.threads.into_inner();
        for (_, thread) in threads {
            self.freeable_kstacks.lock().push(thread.kernel_stack);
        }
        Ok(())
    }

    pub fn current_thread<F: FnOnce(&Thread) -> T, T>(&self, func: F) -> T {
        self.current_process(|process| {
            let threads = process.threads.read();
            func(&threads[self.current_tid().unwrap()])
        })
    }

    pub fn curr_proc_thread<F: FnOnce(&Thread) -> T, T>(
        &self,
        tid: usize,
        func: F,
    ) -> Result<T, InvalidPid> {
        self.current_process(|process| {
            let threads = process.threads.read();
            Ok(func(threads.get(tid).ok_or(InvalidPid)?))
        })
    }

    pub fn current_thread_mut<F: FnOnce(&mut Thread) -> T, T>(&self, func: F) -> T {
        self.current_process_mut(|process| {
            let mut threads = process.threads.write();
            func(&mut threads[self.current_tid().unwrap()])
        })
    }

    pub fn current_process<F: FnOnce(&Process) -> T, T>(&self, func: F) -> T {
        let processes = self.processes.read();
        #[warn(clippy::unwrap_used, reason = "FIXME")]
        #[warn(clippy::indexing_slicing, reason = "FIXME(?)")]
        func(&processes[self.current_pid().unwrap()])
    }

    pub fn current_process_mut<F: FnOnce(&mut Process) -> T, T>(&self, func: F) -> T {
        let mut processes = self.processes.write();
        #[warn(clippy::unwrap_used, reason = "FIXME")]
        #[warn(clippy::indexing_slicing, reason = "FIXME(?)")]
        func(&mut processes[self.current_pid().unwrap()])
    }

    pub fn process<F: FnOnce(&Process) -> T, T>(
        &self,
        pid: usize,
        func: F,
    ) -> Result<T, InvalidPid> {
        let processes = self.processes.read();
        #[warn(clippy::unwrap_used, reason = "FIXME")]
        Ok(func(processes.get(pid).ok_or(InvalidPid)?))
    }

    pub fn process_mut<F: FnOnce(&mut Process) -> T, T>(
        &self,
        pid: usize,
        func: F,
    ) -> Result<T, InvalidPid> {
        let mut processes = self.processes.write();
        #[warn(clippy::unwrap_used, reason = "FIXME")]
        Ok(func(processes.get_mut(pid).ok_or(InvalidPid)?))
    }

    pub fn sleep(&self, reason: SleepReason) {
        #[warn(clippy::unwrap_used, reason = "FIXME")]
        #[warn(clippy::indexing_slicing, reason = "FIXME(?)")]
        self.current_thread(|thread| {
            *thread.sleeping.write() = Some(reason);
        });
        self.task_yield();
    }

    pub fn wake(&self, pid: usize, tid: usize, reason: SleepReason) -> Result<(), InvalidPid> {
        let processes = self.processes.read();
        let process = processes.get(pid).ok_or(InvalidPid)?;
        let threads = process.threads.read();
        let thread = threads.get(tid).ok_or(InvalidPid)?;
        let mut sleeping = thread.sleeping.write();
        if *sleeping == Some(reason) {
            if Some(pid) != self.current_pid() || Some(tid) != self.current_tid() {
                self.ready_to_run.lock().push_back((pid, tid));
            }
            *sleeping = None;
        }
        Ok(())
    }

    pub fn force_wake(&self, pid: usize, tid: usize) -> Result<(), InvalidPid> {
        let processes = self.processes.read();
        let process = processes.get(pid).ok_or(InvalidPid)?;
        let threads = process.threads.read();
        let thread = threads.get(tid).ok_or(InvalidPid)?;
        let mut sleeping = thread.sleeping.write();
        if *sleeping != Some(SleepReason::Exited) {
            // only thing we cannot force a wakeup from
            if Some(pid) != self.current_pid() || Some(tid) != self.current_tid() {
                self.ready_to_run.lock().push_back((pid, tid));
            }
            *sleeping = None;
        }
        Ok(())
    }

    pub fn new_tls_key(&self) -> usize {
        self.current_process_mut(|process| {
            let id = process.num_tls;
            process.num_tls += 1;
            for (_, thread) in &*process.threads().read() {
                thread.tls_values.lock().push(0);
            }
            id
        })
    }

    pub fn set_tls_key(&self, key: usize, val: u64) {
        self.current_thread(|thread| {
            thread.tls_values.lock()[key] = val;
        });
    }

    pub fn get_tls_key(&self, key: usize) -> u64 {
        self.current_thread(|thread| thread.tls_values.lock()[key])
    }

    pub fn new_mutex(&self) -> usize {
        self.current_process(|process| process.mutexes.lock().insert(false))
    }

    pub fn lock_mutex(&self, mutex: usize) {
        loop {
            if self.try_lock_mutex(mutex) {
                break;
            }
            self.sleep(SleepReason::LockedMutex(mutex));
        }
    }

    pub fn lock_mutex_timeout(&self, mutex: usize, duration: u64) -> bool {
        if self.try_lock_mutex(mutex) {
            return true;
        }
        pit::thread_sleeping(self.current_pid().unwrap(), self.current_tid().unwrap(), duration);
        self.sleep(SleepReason::LockedMutex(mutex));
        pit::cancel_sleep(self.current_pid().unwrap(), self.current_tid().unwrap());
        self.try_lock_mutex(mutex)
    }

    pub fn try_lock_mutex(&self, mutex: usize) -> bool {
        self.current_process(|process| {
            let mut mutexes = process.mutexes.lock();
            if !mutexes[mutex] {
                mutexes[mutex] = true;
                true
            } else {
                false
            }
        })
    }

    pub fn unlock_mutex(&self, mutex: usize) {
        self.current_process(|process| {
            let mut mutexes = process.mutexes.lock();
            mutexes[mutex] = false;
            for (tid, thread) in &*process.threads().read() {
                if *thread.sleeping.read() == Some(SleepReason::LockedMutex(mutex)) {
                    self.wake(self.current_pid().unwrap(), tid, SleepReason::LockedMutex(mutex))
                        .unwrap();
                    break;
                }
            }
        });
    }

    pub fn drop_mutex(&self, mutex: usize) {
        self.current_process(|process| process.mutexes.lock().remove(mutex));
    }

    pub fn time_sleep(&self, timeout: u64) {
        pit::thread_sleeping(self.current_pid().unwrap(), self.current_tid().unwrap(), timeout);
        self.sleep(SleepReason::TimeSleep);
    }

    //pub fn print_stats(&self) {
    //    let mut total = KERNEL_SPACE.lock().get_bytes_allocated();
    //    println!(
    //        "[TASKING] Kernel used {}",
    //        SizeFormatter::new(KERNEL_SPACE.lock().get_bytes_allocated(), BINARY)
    //    );
    //    for (i, process) in self.processes.read().iter() {
    //        let bytes_used = process.address_space.as_ref().map_or_else(
    //            || ACTIVE_SPACE.lock().get_bytes_allocated(),
    //            |space| space.get_bytes_allocated(),
    //        );
    //        total += bytes_used;
    //        println!("[TASKING] PID {} used {}", i, SizeFormatter::new(bytes_used, BINARY),);
    //    }
    //    println!("[TASKING] Total used {} ({})", SizeFormatter::new(total, BINARY), total / 4096);
    //}

    pub fn switch_fpu_context(&self) {
        unsafe {
            Cr0::update(|cr0| {
                cr0.remove(Cr0Flags::TASK_SWITCHED);
            });
        }
        let current_fpu_context = *self.current_fpu_context.lock();
        if let Some((fpu_pid, fpu_tid)) = current_fpu_context {
            self.processes.read()[fpu_pid].threads.read()[fpu_tid].fpu_state.lock().save();
        }
        self.current_thread(|thread| thread.fpu_state.lock().load());
        *self.current_fpu_context.lock() =
            Some((self.current_pid().unwrap(), self.current_tid().unwrap()));
    }
}
