use crate::{
    bootinfo::BOOTINFO,
    physical_memory::{PhysicalMemory, PHYSICAL_MEMORY},
};
use alloc::alloc::{AllocError, Allocator, Layout};
use cast::{u64, usize};
use core::{
    fmt,
    ops::Deref,
    ptr::{self, NonNull},
    slice,
    sync::atomic::{AtomicUsize, Ordering},
};
use replace_with::replace_with_or_abort;
use spin::{Lazy, Mutex};
use x86_64::{
    registers::control::Cr3,
    structures::paging::{
        frame::PhysFrameRange,
        mapper::{FlagUpdateError, MapToError, MappedFrame, TranslateResult, UnmapError},
        page::PageRange,
        FrameAllocator, FrameDeallocator, Mapper, OffsetPageTable, Page, PageTable, PageTableFlags,
        PhysFrame, Size4KiB, Translate,
    },
    PhysAddr, VirtAddr,
};

pub struct AddressSpace {
    is_kernel: bool,
    /// This field is used by the IPC code to force the
    /// kernel allocator to allocate user accessible pages
    pub alloc_force_user: bool,
    pub mapper: OffsetPageTable<'static>,
    bytes_allocated: AtomicUsize,
    bump_base: Page,
}

impl fmt::Debug for AddressSpace {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.debug_struct("AddressSpace")
            .field("is_kernel", &self.is_kernel)
            .field("alloc_force_user", &self.alloc_force_user)
            .field("level_4_table", &ptr::from_ref(&self.mapper.level_4_table()))
            .finish_non_exhaustive()
    }
}

#[derive(Debug, Copy, Clone)]
pub enum PagingError {
    RequestInvalid,
    ParentEntryHugePage,
    PageAllocationFailed,
    FrameAllocationFailed,
    PageAlreadyMapped,
    PagesAlreadyMapped,
    PageNotMapped,
    InvalidFrameAddress(
        #[expect(dead_code, reason = "This is useful in debug output when the error is unwrapped")]
        PhysAddr,
    ),
}

impl From<MapToError<Size4KiB>> for PagingError {
    fn from(err: MapToError<Size4KiB>) -> Self {
        match err {
            MapToError::FrameAllocationFailed => Self::FrameAllocationFailed,
            MapToError::ParentEntryHugePage => Self::ParentEntryHugePage,
            MapToError::PageAlreadyMapped(_) => Self::PageAlreadyMapped,
        }
    }
}

impl From<UnmapError> for PagingError {
    fn from(err: UnmapError) -> Self {
        match err {
            UnmapError::ParentEntryHugePage => Self::ParentEntryHugePage,
            UnmapError::PageNotMapped => Self::PageNotMapped,
            UnmapError::InvalidFrameAddress(x) => Self::InvalidFrameAddress(x),
        }
    }
}

impl From<FlagUpdateError> for PagingError {
    fn from(err: FlagUpdateError) -> Self {
        match err {
            FlagUpdateError::PageNotMapped => Self::PageNotMapped,
            FlagUpdateError::ParentEntryHugePage => Self::ParentEntryHugePage,
        }
    }
}

// SAFETY: from_start_address_unchecked requires that the given address is page (4 KiB) aligned,
// which means that the lower 12 bits are zero. We shift the page number left 12 bits, so the lower
// 12 bits will always be zero.
const USER_PAGE_RANGE: PageRange = unsafe {
    PageRange {
        start: Page::from_start_address_unchecked(VirtAddr::new_truncate(1 << 12)),
        end: Page::from_start_address_unchecked(VirtAddr::new_truncate(0x8_0000_0000 << 12)),
    }
};
const KERNEL_PAGE_RANGE: PageRange = unsafe {
    PageRange {
        start: Page::from_start_address_unchecked(VirtAddr::new_truncate(0x8_0000_0000 << 12)),
        end: Page::from_start_address_unchecked(VirtAddr::new_truncate(0xF_FFFF_FFFF << 12)),
    }
};

pub trait AsVirt {
    fn as_virt_ptr<T>(&self) -> *mut T;
    unsafe fn as_virt_ref<'a, T>(&self) -> &'a T {
        unsafe { &*self.as_virt_ptr() }
    }
    unsafe fn as_virt_mut<'a, T>(&self) -> &'a mut T {
        unsafe { &mut *self.as_virt_ptr() }
    }
}

impl AsVirt for PhysFrame {
    fn as_virt_ptr<T>(&self) -> *mut T {
        self.start_address().as_virt_ptr()
    }
}

impl AsVirt for PhysAddr {
    fn as_virt_ptr<T>(&self) -> *mut T {
        #[expect(
            clippy::arithmetic_side_effects,
            reason = "While a sufficiently large amount of physical memory could cause this to wrap, this is so unlikely that paying the cost of checking is not worth it."
        )]
        (*PHYS_OFFSET + self.as_u64()).as_mut_ptr()
    }
}

#[expect(
    clippy::expect_used,
    reason = "Without the physical memory mapping, the kernel cannot function."
)]
pub static PHYS_OFFSET: Lazy<VirtAddr> = Lazy::new(|| {
    VirtAddr::new(
        BOOTINFO
            .physical_memory_offset
            .into_option()
            .expect("Bootloader failed to provide physical memory mapping"),
    )
});

pub struct ASpaceMutex(Mutex<AddressSpace>);

impl ASpaceMutex {
    pub fn new(space: AddressSpace) -> Self {
        Self(Mutex::new(space))
    }
}

impl Deref for ASpaceMutex {
    type Target = Mutex<AddressSpace>;
    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

struct PmemCountingWrapper<'a> {
    pmem: &'a mut PhysicalMemory,
    count: usize,
}

impl<'a> PmemCountingWrapper<'a> {
    fn new(pmem: &'a mut PhysicalMemory) -> Self {
        Self { pmem, count: 0 }
    }
}

unsafe impl FrameAllocator<Size4KiB> for PmemCountingWrapper<'_> {
    #[expect(
        clippy::arithmetic_side_effects,
        reason = "This can never overflow as there are a theoretical maximum of usize::MAX/4096 physical frames."
    )]
    fn allocate_frame(&mut self) -> Option<PhysFrame> {
        if let Some(frame) = self.pmem.allocate_frame() {
            self.count += 1;
            Some(frame)
        } else {
            None
        }
    }
}

pub static KERNEL_SPACE: Lazy<ASpaceMutex> = Lazy::new(|| {
    // SAFETY: Cr3 must point to a valid table otherwise the system would triple fault, so
    // we know it is safe to turn the pointer to it into a reference.
    let table = unsafe { Cr3::read().0.as_virt_mut::<PageTable>() };
    for i in 0..256 {
        table[i].set_addr(PhysAddr::new(0), PageTableFlags::empty());
    }
    for i in 256..512 {
        if table[i].flags().contains(PageTableFlags::PRESENT) {
            let new_flags =
                table[i].flags() | PageTableFlags::WRITABLE | PageTableFlags::USER_ACCESSIBLE;
            table[i].set_flags(new_flags);
        }
    }
    let mut kernel_space = AddressSpace::new_with_addr(table);
    kernel_space.is_kernel = true;
    kernel_space.bump_base = KERNEL_PAGE_RANGE.start;
    let l4_virt = VirtAddr::from_ptr(ptr::from_ref(kernel_space.mapper.level_4_table()));
    #[expect(
        clippy::unwrap_used,
        reason = "This can only fail if the address is invalid. Since the address comes from a reference, that can't happen"
    )]
    let l4_phys = kernel_space.mapper.translate_addr(l4_virt).unwrap();
    unsafe { Cr3::write(PhysFrame::containing_address(l4_phys), Cr3::read().1) };
    ASpaceMutex::new(kernel_space)
});

pub static ACTIVE_SPACE: Lazy<ASpaceMutex> = Lazy::new(|| {
    #[expect(
        clippy::expect_used,
        reason = "If we fail to allocate the first active table, it's fatal."
    )]
    let new_space = AddressSpace::new().expect("Could not allocate new user table");
    let l4_virt = VirtAddr::from_ptr(ptr::from_ref(new_space.mapper.level_4_table()));
    #[expect(
        clippy::unwrap_used,
        reason = "This can only fail if the address is invalid. Since the address comes from a reference, that can't happen"
    )]
    let l4_phys = new_space.mapper.translate_addr(l4_virt).unwrap();
    unsafe { Cr3::write(PhysFrame::containing_address(l4_phys), Cr3::read().1) };
    ASpaceMutex::new(new_space)
});

impl AddressSpace {
    pub fn new() -> Result<Self, PagingError> {
        // SAFETY: We copy the kernel space mappings to the newly allocated table before we make a reference to it, so
        // the reference cannot point to uninitialized data.
        let new_table = unsafe {
            let new_table = PHYSICAL_MEMORY
                .lock()
                .allocate_frame()
                .ok_or(MapToError::FrameAllocationFailed)?
                .as_virt_ptr::<PageTable>();
            new_table.copy_from(KERNEL_SPACE.lock().mapper.level_4_table(), 1);
            &mut *new_table
        };
        let space = Self::new_with_addr(new_table);
        space.record_alloc(4096);
        Ok(space)
    }

    fn new_with_addr(table: &'static mut PageTable) -> Self {
        // SAFETY: The physical offset is passed by the bootloader, so unless the bootloader is
        // buggy (in which case I can't do anything about it), the offset is ok. The requireent
        // about a "valid page table hierarchy" doesn't make much sense, as the only issue I can
        // see, duplicate mappings for a physical frame, is unsafe to make in the first place and
        // so shouldn't have to be guarenteed by us or the direct caller.
        Self {
            mapper: unsafe { OffsetPageTable::new(table, *PHYS_OFFSET) },
            alloc_force_user: false,
            is_kernel: false,
            bytes_allocated: AtomicUsize::new(0),
            bump_base: USER_PAGE_RANGE.start,
        }
    }

    pub fn activate(self) -> Self {
        // This assert should never fire, as the only address space with the is_kernel bool set is
        // stored in the KERNEL_SPACE_INTERNAL static, which is a Mutex, and you can never get
        // ownership of a value in a mutex unless you own the mutex, and no function owns a static.
        assert!(!self.is_kernel);
        Lazy::force(&ACTIVE_SPACE);
        let l4_virt = VirtAddr::from_ptr(ptr::from_ref(self.mapper.level_4_table()));
        #[expect(
            clippy::unwrap_used,
            reason = "This can only fail if the address is invalid. Since the address comes from a reference, that can't happen"
        )]
        let l4_phys = self.mapper.translate_addr(l4_virt).unwrap();
        unsafe { Cr3::write(PhysFrame::containing_address(l4_phys), Cr3::read().1) };
        core::mem::replace(&mut *ACTIVE_SPACE.lock(), self)
    }

    pub fn level_4_table(&self) -> &PageTable {
        self.mapper.level_4_table()
    }

    pub fn level_4_table_mut(&mut self) -> &mut PageTable {
        self.mapper.level_4_table_mut()
    }

    fn check_request_valid(&self, start: Page, length: usize) -> Result<(), PagingError> {
        if (self.is_kernel
            && (start < KERNEL_PAGE_RANGE.start
                || ({
                    #[expect(
                        clippy::arithmetic_side_effects,
                        reason = "The end of the kernel range is the top of the address space. Thus no subtraction from it can underflow."
                    )]
                    let res = length >= usize(KERNEL_PAGE_RANGE.end - start);
                    res
                })))
            || (!self.is_kernel
                && ((start >= USER_PAGE_RANGE.end)
                    || ({
                        #[expect(
                            clippy::arithmetic_side_effects,
                            reason = "The previous check guarenteed the request's start is not greater than the range's end"
                        )]
                        let res = length >= usize(USER_PAGE_RANGE.end - start);
                        res
                    })
                    || {
                        #[expect(
                            clippy::arithmetic_side_effects,
                            reason = "The previous check guarenteed this wouldn't overflow"
                        )]
                        let res = (start + u64(length)) >= USER_PAGE_RANGE.end;
                        res
                    }))
        {
            Err(PagingError::RequestInvalid)
        } else {
            Ok(())
        }
    }

    fn check_request_unmapped(&self, start: Page, length: usize) -> Result<(), PagingError> {
        self.check_request_valid(start, length)?;
        #[expect(
            clippy::arithmetic_side_effects,
            reason = "check_request_valid guarentees this won't overflow"
        )]
        for page in Page::range(start, start + u64(length)) {
            if self.translate_addr(page.start_address()).is_some() {
                return Err(PagingError::PagesAlreadyMapped);
            }
        }
        Ok(())
    }

    /// Runs a provided closure in an address space.
    ///
    /// ## Examples
    /// ```
    /// let address_space = AddressSpace::new();
    /// address_space.run(|| {/* your closure here */});
    /// ```
    pub fn run<F: FnOnce()>(&mut self, func: F) {
        replace_with_or_abort(self, |cl_self| {
            let old_space = cl_self.activate();
            func();
            old_space.activate()
        });
    }

    /// Maps virtual pages and returns the starting address.
    ///
    /// ## Safety
    ///
    /// Creating page table mappings is a fundamentally unsafe operation because
    /// there are various ways to break memory safety through it. For example,
    /// re-mapping an in-use page to a different frame changes and invalidates
    /// all values stored in that page, resulting in undefined behavior on the
    /// next use.
    ///
    /// The caller must ensure that no undefined behavior or memory safety
    /// violations can occur through the new mapping. Among other things, the
    /// caller must prevent the following:
    ///
    /// - Aliasing of `&mut` references, i.e. two `&mut` references that point to
    ///   the same physical address. This is undefined behavior in Rust.
    ///     - This can be ensured by by making sure all frames in the provided range
    ///       are not mapped anywhere else.
    /// - Creating uninitalized or invalid values: Rust requires that all values
    ///   have a correct memory layout. For example, a `bool` must be either a 0
    ///   or a 1 in memory, but not a 3 or 4. An exception is the `MaybeUninit`
    ///   wrapper type, which abstracts over possibly uninitialized memory.
    ///     - This is only a problem when re-mapping pages to different physical
    ///       frames. Mapping pages that are not in use yet is fine.
    unsafe fn map_to(
        &mut self,
        page: Page,
        phys_frame: PhysFrame,
        num_pages: usize,
        flags: PageTableFlags,
    ) -> Result<*mut u8, PagingError> {
        // SAFETY: from_start_address_unchecked requires its argument to be 4k aligned, which it is
        const MAX_PHYS_FRAME: PhysFrame = unsafe {
            PhysFrame::from_start_address_unchecked(PhysAddr::new(0x000F_FFFF_FFFF_F000))
        };
        self.check_request_valid(page, num_pages)?;
        #[expect(
            clippy::arithmetic_side_effects,
            reason = "check_request_valid guarentees this won't overflow"
        )]
        let req_end = page + u64(num_pages);
        if self.bump_base < req_end {
            self.bump_base = req_end;
        }
        #[expect(
            clippy::arithmetic_side_effects,
            reason = "This is the maximum physical frame, so there is no way the subtraction can underflow"
        )]
        if (MAX_PHYS_FRAME - phys_frame) / 4096 < u64(num_pages) {
            return Err(PagingError::RequestInvalid);
        }
        for (page, frame) in (PageRange {
            start: page,
            end: {
                #[expect(
                    clippy::arithmetic_side_effects,
                    reason = "check_request_valid guarentees this won't overflow"
                )]
                let res = page + u64(num_pages);
                res
            },
        })
        .zip(PhysFrameRange {
            start: phys_frame,
            end: {
                #[expect(
                    clippy::arithmetic_side_effects,
                    reason = "The check above guarentees this won't overflow"
                )]
                let res = phys_frame + u64(num_pages);
                res
            },
        }) {
            unsafe {
                let mut phys_mem = PHYSICAL_MEMORY.lock();
                let mut pmem_wrap = PmemCountingWrapper::new(&mut phys_mem);
                self.mapper
                    .map_to_with_table_flags(
                        page,
                        frame,
                        flags | PageTableFlags::PRESENT,
                        PageTableFlags::PRESENT
                            | PageTableFlags::WRITABLE
                            | PageTableFlags::USER_ACCESSIBLE,
                        &mut pmem_wrap,
                    )?
                    .flush();
                #[expect(
                    clippy::arithmetic_side_effects,
                    reason = "This can never overflow, as there are a theoretical maximum of usize::MAX/4096 frames."
                )]
                self.record_alloc(pmem_wrap.count * 4096);
            }
        }
        Ok(page.start_address().as_mut_ptr())
    }

    /// Maps virtual pages to newly allocated physical memory  and returns the starting address.
    /// The newly allocated physical memory contains garbage data, so the mapping will always be writable,
    /// as it doesn't make sense to map garbage data read only.
    ///
    /// ## Safety
    ///
    /// Creating page table mappings is a fundamentally unsafe operation because
    /// there are various ways to break memory safety through it. For example,
    /// re-mapping an in-use page to a different frame changes and invalidates
    /// all values stored in that page, resulting in undefined behavior on the
    /// next use.
    ///
    /// The caller must ensure that no undefined behavior or memory safety
    /// violations can occur through the new mapping by, among other things, preventing
    /// creating uninitalized or invalid values. Rust requires that all values
    /// have a correct memory layout. For example, a `bool` must be either a 0
    /// or a 1 in memory, but not a 3 or 4. An exception is the `MaybeUninit`
    /// wrapper type, which abstracts over possibly uninitialized memory.
    /// Note: You only have to worry about this when re-mapping pages to
    /// different physical frames. Mapping pages that are not in use yet is fine.
    unsafe fn map(
        &mut self,
        page: Page,
        num_pages: usize,
        flags: PageTableFlags,
    ) -> Result<*mut u8, PagingError> {
        self.check_request_valid(page, num_pages)?;
        #[expect(
            clippy::arithmetic_side_effects,
            reason = "check_request_valid guarentees this won't overflow"
        )]
        let req_end = page + u64(num_pages);
        if self.bump_base < req_end {
            self.bump_base = req_end;
        }
        for page in (PageRange { start: page, end: req_end }) {
            unsafe {
                let mut phys_mem = PHYSICAL_MEMORY.lock();
                let frame = phys_mem.allocate_frame().ok_or(PagingError::FrameAllocationFailed)?;
                let mut pmem_wrap = PmemCountingWrapper::new(&mut phys_mem);
                self.mapper
                    .map_to_with_table_flags(
                        page,
                        frame,
                        flags
                            | PageTableFlags::PRESENT
                            | PageTableFlags::WRITABLE
                            | if self.alloc_force_user {
                                PageTableFlags::USER_ACCESSIBLE
                            } else {
                                PageTableFlags::empty()
                            }
                            | PageTableFlags::BIT_9, // Bit 9 is used to indicate the underlying
                        // frame should be freed on process exit
                        PageTableFlags::PRESENT
                            | PageTableFlags::WRITABLE
                            | PageTableFlags::USER_ACCESSIBLE,
                        &mut pmem_wrap,
                    )?
                    .flush();
                #[expect(
                    clippy::arithmetic_side_effects,
                    reason = "This can never overflow, as there are a theoretical maximum of usize::MAX/4096 frames."
                )]
                self.record_alloc(pmem_wrap.count * 4096);
            }
        }
        Ok(page.start_address().as_mut_ptr())
    }

    /// Maps virtual pages to newly allocated contiguous physical memory
    /// and returns the starting address and frame.
    /// The newly allocated physical memory contains garbage data, so the mapping will always be writable,
    /// as it doesn't make sense to map garbage data read only.
    ///
    /// ## Safety
    ///
    /// Creating page table mappings is a fundamentally unsafe operation because
    /// there are various ways to break memory safety through it. For example,
    /// re-mapping an in-use page to a different frame changes and invalidates
    /// all values stored in that page, resulting in undefined behavior on the
    /// next use.
    ///
    /// The caller must ensure that no undefined behavior or memory safety
    /// violations can occur through the new mapping by, among other things, preventing
    /// creating uninitalized or invalid values. Rust requires that all values
    /// have a correct memory layout. For example, a `bool` must be either a 0
    /// or a 1 in memory, but not a 3 or 4. An exception is the `MaybeUninit`
    /// wrapper type, which abstracts over possibly uninitialized memory.
    /// Note: You only have to worry about this when re-mapping pages to
    /// different physical frames. Mapping pages that are not in use yet is fine.
    unsafe fn map_cont_phys(
        &mut self,
        page: Page,
        num_pages: usize,
        flags: PageTableFlags,
    ) -> Result<(*mut u8, u64), PagingError> {
        self.check_request_valid(page, num_pages)?;
        let frame_range = PHYSICAL_MEMORY
            .lock()
            .allocate_frame_range(num_pages)
            .ok_or(PagingError::FrameAllocationFailed)?;
        let phys_start = frame_range.start.start_address().as_u64();
        let res = unsafe {
            self.map_to(page, frame_range.start, num_pages, flags | PageTableFlags::WRITABLE)
                .map(|ptr| (ptr, phys_start))?
        };
        #[expect(
            clippy::arithmetic_side_effects,
            reason = "
            This can never overflow, as there are a theoretical maximum of usize::MAX/4096 pages,
            and check_request_valid makes sure that num_pages is below that max.
        "
        )]
        self.record_alloc(num_pages * 4096);
        Ok(res)
    }

    /// Maps new virtual pages and returns the starting addresss
    ///
    /// ## Safety
    ///
    /// The caller must ensure that no undefined behavior or memory safety
    /// violations can occur through the new mapping by, among other things, preventing
    /// aliasing of `&mut` references, i.e. two `&mut` references that point to the same
    /// physical address. This is undefined behavior in Rust. Aliasing can be prevented
    /// by making sure all frames in the provided range are not mapped anywhere else.
    pub unsafe fn map_free_to(
        &mut self,
        phys_frame: PhysFrame,
        num_pages: usize,
        flags: PageTableFlags,
    ) -> Result<*mut u8, PagingError> {
        let start_page = self.find_free_pages(num_pages)?;
        unsafe { self.map_to(start_page, phys_frame, num_pages, flags) }
    }

    /// Maps new virtual pages to new physical memory and returns the starting address.
    /// The newly allocated physical memory contains garbage data, so the mapping will always be writable,
    /// as it doesn't make sense to map garbage data read only.
    pub fn map_free(
        &mut self,
        num_pages: usize,
        flags: PageTableFlags,
    ) -> Result<*mut u8, PagingError> {
        let start_page = self.find_free_pages(num_pages)?;
        // SAFETY: &mut aliasing is prevented by using free physical frames, and uninitialized
        // values are prevented by using free virtual pages.
        let ptr = unsafe { self.map(start_page, num_pages, flags)? };
        Ok(ptr)
    }

    /// Maps new virtual pages to new contigous physical memory
    /// and returns the starting address and frame.
    /// The newly allocated physical memory contains garbage data, so the mapping will always be writable,
    /// as it doesn't make sense to map garbage data read only.
    pub fn map_free_cont_phys(
        &mut self,
        num_pages: usize,
        flags: PageTableFlags,
    ) -> Result<(*mut u8, u64), PagingError> {
        let start_page = self.find_free_pages(num_pages)?;
        // SAFETY: &mut aliasing is prevented by using free physical frames, and uninitialized
        // values are prevented by using free virtual pages.
        let ptr = unsafe { self.map_cont_phys(start_page, num_pages, flags)? };
        Ok(ptr)
    }

    /// Same behavior as `map`, but asserts that the requested virtual page range is unmapped, and
    /// thus is safe.
    pub fn map_assert_unused(
        &mut self,
        page: Page,
        num_pages: usize,
        flags: PageTableFlags,
    ) -> Result<*mut u8, PagingError> {
        self.check_request_unmapped(page, num_pages)?;
        let ptr = unsafe { self.map(page, num_pages, flags)? };
        Ok(ptr)
    }

    /// Same behavior as `map`, but only maps unmapped pages, and
    /// thus is safe.
    pub fn map_only_unused(
        &mut self,
        page: Page,
        num_pages: usize,
        flags: PageTableFlags,
    ) -> Result<*mut u8, PagingError> {
        #[expect(
            clippy::arithmetic_side_effects,
            reason = "check_request_valid guarentees this won't overflow"
        )]
        let req_end = page + u64(num_pages);
        if self.bump_base < req_end {
            self.bump_base = req_end;
        }
        assert!(!self.alloc_force_user);
        for page in (PageRange { start: page, end: req_end }) {
            if self.translate_addr(page.start_address()).is_some() {
                continue;
            }
            self.record_alloc(4096);
            unsafe {
                let mut phys_mem = PHYSICAL_MEMORY.lock();
                let frame = phys_mem.allocate_frame().ok_or(PagingError::FrameAllocationFailed)?;
                let mut pmem_wrap = PmemCountingWrapper::new(&mut phys_mem);
                self.mapper
                    .map_to_with_table_flags(
                        page,
                        frame,
                        flags
                            | PageTableFlags::PRESENT
                            | PageTableFlags::WRITABLE
                            | PageTableFlags::BIT_9, // Bit 9 is used to indicate the underlying
                        // frame should be freed on process exit
                        PageTableFlags::PRESENT
                            | PageTableFlags::WRITABLE
                            | PageTableFlags::USER_ACCESSIBLE,
                        &mut pmem_wrap,
                    )?
                    .flush();
                #[expect(
                    clippy::arithmetic_side_effects,
                    reason = "This can never overflow, as there are a theoretical maximum of usize::MAX/4096 frames."
                )]
                self.record_alloc(pmem_wrap.count * 4096);
            }
        }
        Ok(page.start_address().as_mut_ptr())
    }

    pub fn unmap(&mut self, page: Page, num_pages: usize) -> Result<(), PagingError> {
        self.check_request_valid(page, num_pages)?;
        #[expect(
            clippy::arithmetic_side_effects,
            reason = "check_request_valid guarentees this won't overflow"
        )]
        for page in (PageRange { start: page, end: page + u64(num_pages) }) {
            unsafe {
                let TranslateResult::Mapped { frame: _, offset: 0, flags } =
                    self.mapper.translate(page.start_address())
                else {
                    continue;
                };
                let (frame, flush) = self.mapper.unmap(page)?;
                if flags.contains(PageTableFlags::BIT_9) {
                    PHYSICAL_MEMORY.lock().deallocate_frame(frame);
                    self.record_dealloc(4096);
                }
                flush.flush();
            }
        }
        Ok(())
    }

    pub fn move_mappings_free(
        &mut self,
        page: Page,
        num_pages: usize,
        dest: &mut Self,
    ) -> Result<Page, PagingError> {
        self.check_request_valid(page, num_pages)?;
        let dest_range_start = dest.find_free_pages(num_pages)?;
        #[expect(
            clippy::arithmetic_side_effects,
            reason = "check_request_valid guarentees this won't overflow"
        )]
        for (src_page, dest_page) in (PageRange { start: page, end: page + u64(num_pages) })
            .zip(PageRange { start: dest_range_start, end: dest_range_start + u64(num_pages) })
        {
            let TranslateResult::Mapped { frame: MappedFrame::Size4KiB(frame), offset: 0, flags } =
                self.mapper.translate(src_page.start_address())
            else {
                return Err(PagingError::PageNotMapped);
            };
            unsafe { dest.map_to(dest_page, frame, 1, flags)? };
            self.mapper.unmap(src_page)?.1.flush();
            if flags.contains(PageTableFlags::BIT_9) {
                self.record_dealloc(4096);
                dest.record_alloc(4096);
            }
        }
        Ok(dest_range_start)
    }

    pub unsafe fn update_flags(
        &mut self,
        page: Page,
        num_pages: usize,
        set_flags: PageTableFlags,
        clear_flags: PageTableFlags,
    ) -> Result<(), PagingError> {
        self.check_request_valid(page, num_pages)?;
        #[expect(
            clippy::arithmetic_side_effects,
            reason = "check_request_valid guarentees this won't overflow"
        )]
        for page in (PageRange { start: page, end: page + u64(num_pages) }) {
            let TranslateResult::Mapped { frame: _, offset: 0, flags } =
                self.mapper.translate(page.start_address())
            else {
                return Err(PagingError::PageNotMapped);
            };
            unsafe {
                self.mapper.update_flags(page, (flags | set_flags) & !(clear_flags))?.flush();
                self.mapper
                    .set_flags_p3_entry(
                        page,
                        PageTableFlags::PRESENT
                            | PageTableFlags::WRITABLE
                            | PageTableFlags::USER_ACCESSIBLE,
                    )?
                    .flush_all();
                self.mapper
                    .set_flags_p2_entry(
                        page,
                        PageTableFlags::PRESENT
                            | PageTableFlags::WRITABLE
                            | PageTableFlags::USER_ACCESSIBLE,
                    )?
                    .flush_all();
            }
        }
        Ok(())
    }

    /// Finds a range of free pages and returns the starting page
    fn find_free_pages(&mut self, num_pages: usize) -> Result<Page, PagingError> {
        if self.is_kernel {
            if num_pages == 0 {
                return Err(PagingError::PageAllocationFailed);
            }
            let mut remaining_pages = num_pages;
            let range = if self.is_kernel { KERNEL_PAGE_RANGE } else { USER_PAGE_RANGE };
            for page in range {
                if self.translate_addr(page.start_address()).is_none() {
                    #[expect(
                        clippy::arithmetic_side_effects,
                        reason = "remaining_pages can never be 0 here, thus this can't underflow"
                    )]
                    {
                        remaining_pages -= 1;
                    }
                    if remaining_pages == 0 {
                        #[expect(
                            clippy::arithmetic_side_effects,
                            reason = "page is at minimum num_pages - 1, thus this can't underflow"
                        )]
                        return Ok(page + 1 - u64(num_pages));
                    }
                } else {
                    remaining_pages = num_pages;
                };
            }
            Err(PagingError::PageAllocationFailed)
        } else {
            #[expect(
                clippy::arithmetic_side_effects,
                reason = "bump_base can never be greater than USER_PAGE_RANGE.end"
            )]
            let num_pages_free = USER_PAGE_RANGE.end - self.bump_base;
            if num_pages_free < u64(num_pages) {
                Err(PagingError::PageAllocationFailed)
            } else {
                let old_base = self.bump_base;
                #[expect(
                    clippy::arithmetic_side_effects,
                    reason = "bump_base can never be greater than USER_PAGE_RANGE.end, and num_pages can never push it over that due to the check above"
                )]
                {
                    self.bump_base += u64(num_pages);
                }
                Ok(old_base)
            }
        }
    }

    pub fn record_alloc(&self, size: usize) {
        self.bytes_allocated.fetch_add(size, Ordering::Relaxed);
    }

    pub fn record_dealloc(&self, size: usize) {
        self.bytes_allocated.fetch_sub(size, Ordering::Relaxed);
    }

    //pub fn get_bytes_allocated(&self) -> usize {
    //    self.bytes_allocated.load(Ordering::Relaxed)
    //}
}

impl Drop for AddressSpace {
    fn drop(&mut self) {
        drop_table(self.mapper.level_4_table(), 4);
    }
}

fn drop_table(table: &PageTable, level: u8) {
    match level.cmp(&1) {
        core::cmp::Ordering::Greater => {
            for (i, entry) in table.iter().enumerate() {
                if level == 4 && i >= 256 {
                    continue;
                }
                if entry.flags().contains(PageTableFlags::PRESENT)
                    && !entry.flags().contains(PageTableFlags::HUGE_PAGE)
                {
                    // SAFETY: The present flag is set on the entry, which means the child frame must
                    // contain a valid page table, so making a reference to it must be ok.
                    #[expect(
                        clippy::unwrap_used,
                        reason = "
                        from_start_address requires it's input to be 4k aligned.
                        The addr method only returns a 4k aligned address if the HUGE_PAGE flag is not set.
                        In addition, the returned address is only valid if the PRESENT flag is set.
                        The if statements has ensured that if we get here, HUGE_FLAG is not set and PRESENT is set.
                    "
                    )]
                    #[expect(
                        clippy::arithmetic_side_effects,
                        reason = "level is at minimum 2 here, thus this can't underflow"
                    )]
                    drop_table(
                        unsafe {
                            PhysFrame::from_start_address(entry.addr()).unwrap().as_virt_ref()
                        },
                        level - 1,
                    );
                }
            }
        }
        core::cmp::Ordering::Equal => {
            for entry in table.iter() {
                if entry.flags().contains(PageTableFlags::PRESENT)
                    && entry.flags().contains(PageTableFlags::BIT_9)
                {
                    #[expect(
                        clippy::unwrap_used,
                        reason = "This fails if the entry isn't present or it's a huge page. Presence was verified above, and level 1 tables cannot have huge entries."
                    )]
                    unsafe {
                        PHYSICAL_MEMORY.lock().deallocate_frame(entry.frame().unwrap());
                    };
                }
            }
        }
        core::cmp::Ordering::Less => unreachable!(),
    }
    #[expect(
        clippy::unwrap_used,
        reason = "This can only fail if the address is invalid. Since the address comes from a reference, that can't happen"
    )]
    let phys_addr = ACTIVE_SPACE.lock().translate_addr(VirtAddr::from_ptr(table)).unwrap();
    #[expect(
        clippy::unwrap_used,
        reason = "PhysFrame requires its argument to be 4k aligned, which is guaranteed since page tables must be 4k aligned"
    )]
    unsafe {
        PHYSICAL_MEMORY.lock().deallocate_frame(PhysFrame::from_start_address(phys_addr).unwrap());
    };
}

impl Translate for AddressSpace {
    fn translate(&self, addr: VirtAddr) -> TranslateResult {
        self.mapper.translate(addr)
    }
}

unsafe impl Allocator for ASpaceMutex {
    fn allocate(&self, layout: Layout) -> Result<NonNull<[u8]>, AllocError> {
        if layout.align() > 4096 {
            return Err(AllocError);
        }
        let size = layout.size().next_multiple_of(4096);
        let mut space = self.0.lock();
        let flags = if !space.is_kernel || space.alloc_force_user {
            PageTableFlags::USER_ACCESSIBLE
        } else {
            PageTableFlags::empty()
        };
        let start = space.map_free(size / 4096, flags).map_err(|_| AllocError)?;
        Ok(unsafe { slice::from_raw_parts_mut(start.cast::<u8>(), size) }.into())
    }

    unsafe fn deallocate(&self, ptr: NonNull<u8>, layout: Layout) {
        #[expect(
            clippy::unwrap_used,
            reason = "
            from_start_address requires it's input to be 4k aligned.
            Calling deallocate requires that the pointer was returned by the same allocator.
            The allocator only returns 4k aligned pointers, so from_start_address cannot panic.
        "
        )]
        let start_page =
            Page::from_start_address(VirtAddr::new(u64(ptr.as_ptr().expose_provenance()))).unwrap();
        let length = layout.size().div_ceil(4096);
        #[expect(
            clippy::unwrap_used,
            reason = "
            Unmap only fails if a subpage of a huge page is freed, or if the mapping is invalid.
            The kernel doesn't use huge pages and the mapping must be valid to be returned from allocate, so unmap cannot fail.
        "
        )]
        self.0.lock().unmap(start_page, length).unwrap();
    }
}
